# Licensed under a 3-clause BSD style license - see LICENSE.rst

import fnmatch
import logging
import re
import warnings
from collections import OrderedDict
from os import listdir, path

import astropy.io.fits as fits
import numpy as np
import numpy.ma as ma
from astropy.table import MaskedColumn, Table
from astropy.utils.exceptions import AstropyUserWarning

from .ccddata import _recognized_fits_file_extensions, fits_ccddata_reader

logger = logging.getLogger(__name__)

__all__ = ["ImageFileCollection"]
__doctest_skip__ = ["*"]


class ImageFileCollection:
    """
    Representation of a collection of image files.

    The class offers a table summarizing values of
    keywords in the FITS headers of the files in the collection and offers
    convenient methods for iterating over the files in the collection. The
    generator methods use simple filtering syntax and can automate storage
    of any FITS files modified in the loop using the generator.

    Parameters
    ----------
    location : str or None, optional
        Path to directory containing FITS files.
        Default is ``None``.

    keywords : list of str, '*' or None, optional
        Keywords that should be used as column headings in the summary table.
        If the value is or includes '*' then all keywords that appear in any
        of the FITS headers of the files in the collection become table
        columns. Default value is '*' unless ``info_file`` is specified.
        Default is ``None``.

    find_fits_by_reading: bool, optional
        If ``True``, read each file in location to check whether the file is a
        FITS file and include it in the collection based on that, rather than
        by file name. Compressed files, e.g. image.fits.gz, will **NOT** be
        properly detected. *Will be ignored if `filenames` is not ``None``.*

    filenames: str, list of str, or None, optional
        List of the names of FITS files which will be added to the collection.
        The filenames may either be in ``location`` or the name can be a
        relative or absolute path to the file.
        Default is ``None``.

    glob_include: str or None, optional
        Unix-style filename pattern to select filenames to include in the file
        collection. Can be used in conjunction with ``glob_exclude`` to
        easily select subsets of files in the target directory.
        Default is ``None``.

    glob_exclude: str or None, optional
        Unix-style filename pattern to select filenames to exclude from the
        file collection. Can be used in conjunction with ``glob_include`` to
        easily select subsets of files in the target directory.
        Default is ``None``.

    ext: str or int, optional
        The extension from which the header and data will be read in all
        files.Default is ``0``.

    Raises
    ------
    ValueError
        Raised if keywords are set to a combination of '*' and any other
        value.
    """

    def __init__(
        self,
        location=None,
        keywords=None,
        find_fits_by_reading=False,
        filenames=None,
        glob_include=None,
        glob_exclude=None,
        ext=0,
    ):
        # Include or exclude files from the collection based on glob pattern
        # matching - has to go above call to _get_files()
        if glob_exclude is not None:
            glob_exclude = str(glob_exclude)  # some minimal validation
        self._glob_exclude = glob_exclude

        if glob_include is not None:
            glob_include = str(glob_include)
        self._glob_include = glob_include

        if location is not None:
            self._location = location
        else:
            self._location = ""

        self._find_fits_by_reading = find_fits_by_reading

        self._filenames = filenames
        self._files = []
        self._files = self._get_files()

        if self._files == []:
            warnings.warn(
                "no FITS files in the collection.", AstropyUserWarning, stacklevel=2
            )
        self._summary = {}
        if keywords is None:
            # Use all keywords.
            keywords = "*"

        # Used internally to keep track of whether the user asked for all
        # keywords or a specific list. The keywords setter takes care of
        # actually setting the correct value, this just ensure that there
        # is always *some* value.
        self._all_keywords = False

        self._ext = ext

        if keywords:
            self.keywords = keywords

    def __repr__(self):
        if self.location is None:
            location = ""
        else:
            location = f"location={self.location!r}"

        if self._all_keywords:
            kw = ""
        else:
            kw = f"keywords={self.keywords[1:]!r}"

        if self.glob_exclude is None:
            glob_exclude = ""
        else:
            glob_exclude = f"glob_exclude={self.glob_exclude!r}"

        if self.glob_include is None:
            glob_include = ""
        else:
            glob_include = f"glob_include={self.glob_include!r}"

        if self.ext == 0:
            ext = ""
        else:
            ext = f"ext={self.ext}"

        if self._filenames is None:
            filenames = ""
        else:
            filenames = f"filenames={self._filenames}"

        params = [location, kw, filenames, glob_include, glob_exclude, ext]
        params = ", ".join([p for p in params if p])

        str_repr = f"{self.__class__.__name__}({params})"

        return str_repr

    @property
    def summary(self):
        """
        `~astropy.table.Table` of values of FITS keywords for files in the
        collection.

        Each keyword is a column heading. In addition, there is a column
        called ``file`` that contains the name of the FITS file. The directory
        is not included as part of that name.

        The first column is always named ``file``.

        The order of the remaining columns depends on how the summary was
        constructed.

        If a wildcard, ``*`` was used then the order is the order in which
        the keywords appear in the FITS files from which the summary is
        constructed.

        If an explicit list of keywords was supplied in setting up the
        collection then the order of the columns is the order of the
        keywords.
        """
        return self._summary

    @property
    def location(self):
        """
        str, Path name to directory containing FITS files.
        """
        return self._location

    @property
    def keywords(self):
        """
        list of str, Keywords currently in the summary table.

        Setting the keywords causes the summary table to be regenerated unless
        the new keywords are a subset of the old.

        .. versionchanged:: 1.3
            Added ``deleter`` for ``keywords`` property.
        """
        if self.summary:
            return self.summary.keys()
        else:
            return []

    @keywords.setter
    def keywords(self, keywords):
        # since keywords are drawn from self.summary, setting
        # summary sets the keywords.
        if keywords is None:
            self._summary = []
            return

        if keywords == "*":
            self._all_keywords = True
        else:
            self._all_keywords = False

        logging.debug("keywords in setter before pruning: %s.", keywords)

        # remove duplicates and force a copy so we can sort the items later
        # by their given position.
        new_keys_set = set(keywords)
        new_keys_lst = list(new_keys_set)
        new_keys_set.add("file")

        logging.debug("keywords after pruning %s.", new_keys_lst)

        current_set = set(self.keywords)
        if new_keys_set.issubset(current_set):
            logging.debug("table columns before trimming: %s.", " ".join(current_set))
            cut_keys = current_set.difference(new_keys_set)
            logging.debug("will try removing columns: %s.", " ".join(cut_keys))
            for key in cut_keys:
                self._summary.remove_column(key)
            logging.debug(
                "after removal column names are: %s.", " ".join(self.keywords)
            )
        else:
            logging.debug("should be building new table...")
            # Reorder the keywords to match the initial ordering.
            new_keys_lst.sort(key=keywords.index)
            self._summary = self._fits_summary(new_keys_lst)

    @keywords.deleter
    def keywords(self):
        # since keywords are drawn from self._summary, setting
        # _summary = [] deletes the keywords.
        self._summary = []

    @property
    def files(self):
        """
        list of str, Unfiltered list of FITS files in location.
        """
        return self._files

    @property
    def glob_include(self):
        """
        str or None, Unix-style filename pattern to select filenames to include
        in the file collection.
        """
        return self._glob_include

    @property
    def glob_exclude(self):
        """
        str or None, Unix-style filename pattern to select filenames to exclude
        in the file collection.
        """
        return self._glob_exclude

    @property
    def ext(self):
        """
        str or int, The extension from which the header and data will
        be read in all files.
        """
        return self._ext

    def values(self, keyword, unique=False):
        """
        List of values for a keyword.

        Parameters
        ----------
        keyword : str
            Keyword (i.e. table column) for which values are desired.

        unique : bool, optional
            If True, return only the unique values for the keyword.
            Default is ``False``.

        Returns
        -------
        list
            Values as a list.
        """
        if keyword not in self.keywords:
            raise ValueError(f"keyword {keyword} is not in the current summary")
        if unique:
            return list(set(self.summary[keyword].tolist()))
        else:
            return self.summary[keyword].tolist()

    def files_filtered(self, **kwd):
        """Determine files whose keywords have listed values.

        Parameters
        ----------
        include_path : bool, keyword-only
            If the keyword ``include_path=True`` is set, the returned list
            contains not just the filename, but the full path to each file.
            Default is ``False``.

        regex_match : bool, keyword-only
            If ``True``, then string values in the ``**kwd`` dictionary are
            treated as regular expression patterns and matching is done by
            regular expression search. The search is always
            **case insensitive**.

        **kwd :
            ``**kwd`` is dict of keywords and values the files must have.
            The value '*' represents any value.
            A missing keyword is indicated by value ''.

        Returns
        -------
        filenames : list
            The files that satisfy the keyword-value restrictions specified by
            the ``**kwd``.

        Examples
        --------
        Some examples for filtering::

            >>> keys = ['imagetyp','filter']
            >>> collection = ImageFileCollection('test/data', keywords=keys)
            >>> collection.files_filtered(imagetyp='LIGHT', filter='R')
            >>> collection.files_filtered(imagetyp='*', filter='')

        In case you want to filter with keyword names that cannot be used
        as keyword argument name, you have to unpack them using a dictionary.
        For example if a keyword name contains a space or a ``-``::

            >>> add_filters = {'exp-time': 20, 'ESO TPL ID': 1050}
            >>> collection.files_filtered(imagetyp='LIGHT', **add_filters)

        Notes
        -----
        Value comparison is case *insensitive* for strings, whether matching
        exactly or matching with regular expressions.
        """
        # If the collection is empty, self.summary == None; return empty list
        if self.summary is None:
            return []

        # force a copy by explicitly converting to a list
        current_file_mask = self.summary["file"].mask.tolist()

        include_path = kwd.pop("include_path", False)

        self._find_keywords_by_values(**kwd)
        filtered_files = self.summary["file"].compressed()
        self.summary["file"].mask = current_file_mask
        if include_path:
            filtered_files = [
                path.join(self._location, f) for f in filtered_files.tolist()
            ]
        return filtered_files

    def refresh(self):
        """
        Refresh the collection by re-reading headers.
        """
        keywords = "*" if self._all_keywords else self.keywords
        # Re-load list of files
        self._files = self._get_files()
        self._summary = self._fits_summary(header_keywords=keywords)

    def sort(self, keys):
        """Sort the list of files to determine the order of iteration.

        Sort the table of files according to one or more keys. This does not
        create a new object, instead is sorts in place.

        Parameters
        ----------
        keys : str, list of str
            The key(s) to order the table by.
        """
        if len(self._summary) > 0:
            self._summary.sort(keys)
            self._files = self.summary["file"].tolist()

    def filter(self, **kwd):
        """
        Create a new collection by filtering the current collection.

        Parameters
        ----------

        regex_match : bool, keyword-only
            If ``True``, then string values in the ``**kwd`` dictionary are
            treated as regular expression patterns and matching is done by
            regular expression search. The search is always
            **case insensitive**.

        **kwd :
            ``**kwd`` is dict of keywords and values the files must have.
            The value '*' represents any value.
            A missing keyword is indicated by value ''.

        Returns
        -------

        `ImageFileCollection`
            A new collection with the files matched by the arguments
            to filter.
        """
        files = self.files_filtered(include_path=True, **kwd)
        return ImageFileCollection(filenames=files, keywords=self.keywords)

    def _get_files(self):
        """Helper method which checks whether ``files`` should be set
        to a subset of file names or to all file names in a directory.

        Returns
        -------
        files : list or str
            List of file names which will be added to the collection.
        """
        files = []
        if self._filenames:
            if isinstance(self._filenames, str):
                files.append(self._filenames)
            else:
                files = self._filenames
        else:
            # Check if self.location is set, otherwise proceed with empty list
            if self.location != "":
                files = self._fits_files_in_directory()

            if self.glob_include is not None:
                files = fnmatch.filter(files, self.glob_include)
            if self.glob_exclude is not None:
                files = [
                    file
                    for file in files
                    if not fnmatch.fnmatch(file, self.glob_exclude)
                ]

        return files

    def _dict_from_fits_header(
        self, file_name, input_summary=None, missing_marker=None
    ):
        """
        Construct an ordered dictionary whose keys are the header keywords
        and values are a list of the values from this file and the input
        dictionary. If the input dictionary is ordered then that order is
        preserved.

        Parameters
        ----------
        file_name : str
            Name of FITS file.

        input_summary : dict or None, optional
            Existing dictionary to which new values should be appended.
            Default is ``None``.

        missing_marker : any type, optional
            Fill value for missing header-keywords.
            Default is ``None``.

        Returns
        -------
        file_table : `~astropy.table.Table`
        """

        def _add_val_to_dict(key, value, tbl_dict, n_previous, missing_marker):
            try:
                tbl_dict[key].append(value)
            except KeyError:
                tbl_dict[key] = [missing_marker] * n_previous
                tbl_dict[key].append(value)

        if input_summary is None:
            summary = OrderedDict()
            n_previous = 0
        else:
            summary = input_summary
            n_previous = len(summary["file"])

        h = fits.getheader(file_name, self.ext)

        assert "file" not in h

        if self.location:
            # We have a location and can reconstruct the path using it
            name_for_file_column = path.basename(file_name)
        else:
            # No location, so use whatever path the user passed in
            name_for_file_column = file_name

        # Try opening header before this so that file name is only added if
        # file is valid FITS
        try:
            summary["file"].append(name_for_file_column)
        except KeyError:
            summary["file"] = [name_for_file_column]

        missing_in_this_file = [k for k in summary if (k not in h and k != "file")]

        multi_entry_keys = {"comment": [], "history": []}

        alreadyencountered = set()
        for k, v in h.items():
            if k == "":
                continue

            k = k.lower()

            if k in ["comment", "history"]:
                multi_entry_keys[k].append(str(v))
                # Accumulate these in a separate dictionary until the
                # end to avoid adding multiple entries to summary.
                continue
            elif k in alreadyencountered:
                # The "normal" multi-entries HISTORY, COMMENT and BLANK are
                # already processed so any further duplication is probably
                # a mistake. It would lead to problems in ImageFileCollection
                # to add it as well, so simply ignore those.
                warnings.warn(
                    f'Header from file "{file_name}" contains multiple entries for '
                    f'"{k}", the pair "{k}={v}" will be ignored.',
                    UserWarning,
                    stacklevel=2,
                )
                continue
            else:
                # Add the key to the already encountered keys so we don't add
                # it more than once.
                alreadyencountered.add(k)

            _add_val_to_dict(k, v, summary, n_previous, missing_marker)

        for k, v in multi_entry_keys.items():
            if v:
                joined = ",".join(v)
                _add_val_to_dict(k, joined, summary, n_previous, missing_marker)

        for missing in missing_in_this_file:
            summary[missing].append(missing_marker)

        return summary

    def _set_column_name_case_to_match_keywords(self, header_keys, summary_table):
        for k in header_keys:
            k_lower = k.lower()
            if k_lower != k:
                try:
                    summary_table.rename_column(k_lower, k)
                except KeyError:
                    pass

    def _fits_summary(self, header_keywords):
        """
        Generate a summary table of keywords from FITS headers.

        Parameters
        ----------
        header_keywords : list of str or '*'
            Keywords whose value should be extracted from FITS headers or '*'
            to extract all.
        """

        if not self.files:
            return None

        # Make sure we have a list...for example, in python 3, dict.keys()
        # is not a list.
        original_keywords = list(header_keywords)

        # Get rid of any duplicate keywords, also forces a copy.
        header_keys = set(original_keywords)
        header_keys.add("file")

        file_name_column = MaskedColumn(name="file", data=self.files)

        if not header_keys or (header_keys == {"file"}):
            summary_table = Table(masked=True)
            summary_table.add_column(file_name_column)
            return summary_table

        summary_dict = None
        missing_marker = None

        for file_name in file_name_column.tolist():
            file_path = path.join(self.location, file_name)
            try:
                # Note: summary_dict is an OrderedDict, so should preserve
                # the order of the keywords in the FITS header.
                summary_dict = self._dict_from_fits_header(
                    file_path, input_summary=summary_dict, missing_marker=missing_marker
                )
            except OSError as e:
                logger.warning(
                    "unable to get FITS header for file %s: %s.", file_path, e
                )
                continue

        summary_table = Table(summary_dict, masked=True)

        for column in summary_table.colnames:
            summary_table[column].mask = [
                v is missing_marker for v in summary_table[column].tolist()
            ]

        self._set_column_name_case_to_match_keywords(header_keys, summary_table)
        missing_columns = header_keys - set(summary_table.colnames)
        missing_columns -= {"*"}

        length = len(summary_table)
        for column in missing_columns:
            all_masked = MaskedColumn(
                name=column, data=np.zeros(length), mask=np.ones(length)
            )
            summary_table.add_column(all_masked)

        if "*" not in header_keys:
            # Rearrange table columns to match order of keywords.
            # File always comes first.
            header_keys -= {"file"}
            original_order = ["file"] + sorted(header_keys, key=original_keywords.index)
            summary_table = summary_table[original_order]

        if not summary_table.masked:
            summary_table = Table(summary_table, masked=True)

        return summary_table

    def _find_keywords_by_values(self, **kwd):
        """
        Find files whose keywords have given values.

        Parameters
        ----------

        match_regex : bool, optional
            If ``True`` match string values by using a regular expression
            search instead of equality. Default value is ``False``.

        The remaining arguments are keyword/value pairs specifying the
        values to match.

        `**kwd` is list of keywords and values the files must have.

        The value '*' represents any value.
        A missing keyword is indicated by value ''

        Example::

            >>> keys = ['imagetyp','filter']
            >>> collection = ImageFileCollection('test/data', keywords=keys)
            >>> collection.files_filtered(imagetyp='LIGHT', filter='R')
            >>> collection.files_filtered(imagetyp='*', filter='')
            >>> collection.files_filtered(imagetyp='bias|filter', regex_match=True)

        NOTE: Value comparison is case *insensitive* for strings.
        """
        regex_match = kwd.pop("regex_match", False)
        keywords = kwd.keys()
        values = kwd.values()

        if set(keywords).issubset(self.keywords):
            # we already have the information in memory
            use_info = self.summary
        else:
            # we need to load information about these keywords.
            use_info = self._fits_summary(header_keywords=keywords)

        matches = np.ones(len(use_info), dtype=bool)
        for key, value in zip(keywords, values):
            logger.debug("key %s, value %s", key, value)
            logger.debug("value in table %s", use_info[key])
            value_missing = use_info[key].mask
            logger.debug("value missing: %s", value_missing)
            value_not_missing = np.logical_not(value_missing)
            if value == "*":
                have_this_value = value_not_missing
            elif value is not None:
                if isinstance(value, str):
                    # need to loop explicitly over array rather than using
                    # where to correctly do string comparison.
                    have_this_value = np.zeros(len(use_info), dtype=bool)

                    # We are going to do a regex match no matter what.
                    if regex_match:
                        pattern = re.compile(value, flags=re.IGNORECASE)
                    else:
                        # Escape all special characters that might be present
                        value = re.escape(value)
                        # This pattern matches the prior behavior.
                        pattern = re.compile("^" + value + "$", flags=re.IGNORECASE)

                    for idx, file_key_value in enumerate(use_info[key].tolist()):
                        if value_not_missing[idx]:
                            try:
                                value_matches = (
                                    pattern.search(file_key_value) is not None
                                )
                            except TypeError:
                                # In case we're dealing with an object column
                                # there could be values other than strings in it
                                # so it could fail with an TypeError.
                                value_matches = False
                        else:
                            value_matches = False

                        have_this_value[idx] = value_not_missing[idx] & value_matches
                else:
                    have_this_value = value_not_missing
                    tmp = use_info[key][value_not_missing] == value
                    have_this_value[value_not_missing] = tmp
                    have_this_value &= value_not_missing
            else:
                # this case--when value==None--is asking for the files which
                # are missing a value for this keyword
                have_this_value = value_missing

            matches &= have_this_value

        # the numpy convention is that the mask is True for values to
        # be omitted, hence use ~matches.
        logger.debug("Matches: %s", matches)
        self.summary["file"].mask = ma.nomask
        self.summary["file"].mask[~matches] = True

    def _fits_files_in_directory(self, extensions=None, compressed=True):
        """
        Get names of FITS files in directory, based on filename extension.

        Parameters
        ----------
        extensions : list of str or None, optional
            List of filename extensions that are FITS files. Default is
            ``['fit', 'fits', 'fts']``.
            Default is ``None``.

        compressed : bool, optional
            If ``True``, compressed files should be included in the list
            (e.g. `.fits.gz`).
            Default is ``True``.

        Returns
        -------
        list
            *Names* of the files (with extension), not the full pathname.
        """

        full_extensions = extensions or list(_recognized_fits_file_extensions)

        # The common compressed fits image .fz is supported using ext=1 when
        # calling ImageFileCollection
        if compressed:
            for comp in [".gz", ".bz2", ".Z", ".zip", ".fz"]:
                with_comp = [extension + comp for extension in full_extensions]
                full_extensions.extend(with_comp)

        all_files = listdir(self.location)
        files = []
        if not self._find_fits_by_reading:
            for extension in full_extensions:
                files.extend(fnmatch.filter(all_files, "*" + extension))
        else:
            for infile in all_files:
                inpath = path.join(self.location, infile)
                with open(inpath, "rb") as fp:
                    # Hmm, first argument to is_fits is not actually used in
                    # that function. *shrug*
                    if fits.connect.is_fits("just some junk", infile, fp):
                        files.append(infile)

        files.sort()
        return files

    def _generator(
        self,
        return_type,
        save_with_name="",
        save_location="",
        overwrite=False,
        do_not_scale_image_data=True,
        return_fname=False,
        ccd_kwargs=None,
        **kwd,
    ):
        """
        Generator that yields each {name} in the collection.

        If any of the parameters ``save_with_name``, ``save_location`` or
        ``overwrite`` evaluates to ``True`` the generator will write a copy of
        each FITS file it is iterating over. In other words, if
        ``save_with_name`` and/or ``save_location`` is a string with non-zero
        length, and/or ``overwrite`` is ``True``, a copy of each FITS file will
        be made.

        Parameters
        ----------
        save_with_name : str, optional
            string added to end of file name (before extension) if
            FITS file should be saved after iteration. Unless
            ``save_location`` is set, files will be saved to location of
            the source files ``self.location``.
            Default is ``''``.

        save_location : str, optional
            Directory in which to save FITS files; implies that FITS
            files will be saved. Note this provides an easy way to
            copy a directory of files--loop over the {name} with
            ``save_location`` set.
            Default is ``''``.

        overwrite : bool, optional
            If ``True``, overwrite input FITS files.
            Default is ``False``.

        do_not_scale_image_data : bool, optional
            If ``True``, prevents fits from scaling images. Default is
            ``{default_scaling}``.
            Default is ``True``.

        return_fname : bool, optional
            If True, return the tuple (header, file_name) instead of just
            header. The file name returned is the name of the file only,
            not the full path to the file.
            Default is ``False``.

        ccd_kwargs : dict, optional
            Dict with parameters for `~astropy.nddata.fits_ccddata_reader`.
            For instance, the key ``'unit'`` can be used to specify the unit
            of the data. If ``'unit'`` is not given then ``'adu'`` is used as
            the default unit.
            See `~astropy.nddata.fits_ccddata_reader` for a complete list of
            parameters that can be passed through ``ccd_kwargs``.


        regex_match : bool, keyword-only
            If ``True``, then string values in the ``**kwd`` dictionary are
            treated as regular expression patterns and matching is done by
            regular expression search. The search is always
            **case insensitive**.

        **kwd :
            Any additional keywords are used to filter the items returned; see
            `files_filtered` examples for details.

        Returns
        -------
        `{return_type}`
            If ``return_fname`` is ``False``, yield the next {name} in the
            collection.

        (`{return_type}`, str)
            If ``return_fname`` is ``True``, yield a tuple of
            ({name}, ``file name``) for the next item in the collection.
        """
        # store mask so we can reset at end--must COPY, otherwise
        # current_mask just points to the mask of summary
        if not self.summary:
            return

        current_mask = {}
        for col in self.summary.columns:
            current_mask[col] = self.summary[col].mask

        if kwd:
            self._find_keywords_by_values(**kwd)

        ccd_kwargs = ccd_kwargs or {}

        for full_path in self._paths():
            add_kwargs = {"do_not_scale_image_data": do_not_scale_image_data}

            # We need to open the file here, get the appropriate values and then
            # close it again before it "yields" otherwise it's not garantueed
            # that the generator actually advances and closes the file again.
            # For example if one uses "next" on the generator manually the
            # file handle could "leak".
            if return_type == "header":
                return_thing = fits.getheader(full_path, self.ext)
            elif return_type == "data":
                return_thing = fits.getdata(full_path, self.ext, **add_kwargs)
            elif return_type == "ccd":
                return_thing = fits_ccddata_reader(
                    full_path, hdu=self.ext, **ccd_kwargs
                )
            elif return_type == "hdu":
                with fits.open(full_path, **add_kwargs) as hdulist:
                    ext_index = hdulist.index_of(self.ext)
                    # Need to copy the HDU to prevent lazy loading problems
                    # and "IO operations on closed file" errors
                    return_thing = hdulist[ext_index].copy()
            else:
                raise ValueError(f"no generator for {return_type}")

            file_name = path.basename(full_path)
            if return_fname:
                yield return_thing, file_name
            else:
                yield return_thing

            if save_location:
                destination_dir = save_location
            else:
                destination_dir = path.dirname(full_path)
            basename = path.basename(full_path)
            if save_with_name:
                base, ext = path.splitext(basename)
                basename = base + save_with_name + ext

            new_path = path.join(destination_dir, basename)

            if return_type == "ccd":
                pass
            elif (new_path != full_path) or overwrite:
                with fits.open(full_path, **add_kwargs) as hdulist:
                    ext_index = hdulist.index_of(self.ext)
                    if return_type == "hdu":
                        hdulist[ext_index] = return_thing
                    elif return_type == "data":
                        hdulist[ext_index].data = return_thing
                    elif return_type == "header":
                        hdulist[ext_index].header = return_thing

                    try:
                        hdulist.writeto(new_path, overwrite=overwrite)
                    except OSError:
                        logger.error("error writing file %s", new_path)
                        raise

        # reset mask
        for col in self.summary.columns:
            self.summary[col].mask = current_mask[col]

    def _paths(self):
        """
        Full path to each file.
        """
        unmasked_files = self.summary["file"].compressed().tolist()
        return [path.join(self.location, file_) for file_ in unmasked_files]

    def headers(self, do_not_scale_image_data=True, **kwd):
        return self._generator(
            "header", do_not_scale_image_data=do_not_scale_image_data, **kwd
        )

    headers.__doc__ = _generator.__doc__.format(
        name="header", default_scaling="True", return_type="astropy.io.fits.Header"
    )

    def hdus(self, do_not_scale_image_data=False, **kwd):
        return self._generator(
            "hdu", do_not_scale_image_data=do_not_scale_image_data, **kwd
        )

    hdus.__doc__ = _generator.__doc__.format(
        name="HDU",
        default_scaling="False",
        return_type="`, ` ".join(
            ("astropy.io.fits.PrimaryHDU", "astropy.io.fits.ImageHDU")
        ),
    )

    def data(self, do_not_scale_image_data=False, **kwd):
        return self._generator(
            "data", do_not_scale_image_data=do_not_scale_image_data, **kwd
        )

    data.__doc__ = _generator.__doc__.format(
        name="image", default_scaling="False", return_type="numpy.ndarray"
    )

    def ccds(self, ccd_kwargs=None, **kwd):
        if (clobber := kwd.get("clobber")) is not None:
            warnings.warn(
                "The 'clobber' keyword argument is a deprecated alias for 'overwrite'",
                category=DeprecationWarning,
                stacklevel=2,
            )
            kwd["overwrite"] = clobber
        if kwd.get("overwrite"):
            raise NotImplementedError("overwrite=True is not supported for CCDs.")
        return self._generator("ccd", ccd_kwargs=ccd_kwargs, **kwd)

    ccds.__doc__ = _generator.__doc__.format(
        name="CCDData", default_scaling="True", return_type="astropy.nddata.CCDData"
    )



import gzip
import os
from tempfile import mkdtemp

import numpy as np
from astropy.io import fits


def _make_file_for_testing(file_name="", **kwd):
    img = np.uint16(np.arange(100))

    hdu = fits.PrimaryHDU(img)

    for k, v in kwd.items():
        hdu.header[k] = v

    hdu.writeto(file_name)


def directory_for_testing():
    """
    Set up directory with these contents:

    One file with imagetyp BIAS. It has an the keyword EXPOSURE in
    the header, but no others beyond IMAGETYP and the bare minimum
    created with the FITS file.

    File name(s)
    ------------

    no_filter_no_object_bias.fit

    Five (5) files with imagetyp LIGHT, including two compressed
    files.

    + One file for each compression type, currently .gz and .fz.
    + ALL of the files will have the keyword EXPOSURE
      in the header.
    + Only ONE of them will have the value EXPOSURE=15.0.
    + All of the files EXCEPT ONE will have the keyword
      FILTER with the value 'R'.
    + NONE of the files have the keyword OBJECT

    File names
    ----------

    test.fits.fz
    filter_no_object_light.fit
    filter_object_light.fit.gz
    filter_object_light.fit
    no_filter_no_object_light.fit    <---- this one has no filter
    """
    n_test = {
        "files": 6,
        "missing_filter_value": 1,
        "bias": 1,
        "compressed": 2,
        "light": 5,
    }

    test_dir = mkdtemp()

    # Directory is reset on teardown.
    original_dir = os.getcwd()
    os.chdir(test_dir)

    _make_file_for_testing(
        file_name="no_filter_no_object_bias.fit", imagetyp="BIAS", EXPOSURE=0.0
    )

    _make_file_for_testing(
        file_name="no_filter_no_object_light.fit", imagetyp="LIGHT", EXPOSURE=1.0
    )

    _make_file_for_testing(
        file_name="filter_no_object_light.fit",
        imagetyp="LIGHT",
        EXPOSURE=1.0,
        filter="R",
    )

    _make_file_for_testing(
        file_name="filter_object_light.fit", imagetyp="LIGHT", EXPOSURE=1.0, filter="R"
    )

    with open("filter_object_light.fit", "rb") as f_in:
        with gzip.open("filter_object_light.fit.gz", "wb") as f_out:
            f_out.write(f_in.read())

    # filter_object.writeto('filter_object_RA_keyword_light.fit')

    _make_file_for_testing(
        file_name="test.fits.fz", imagetyp="LIGHT", EXPOSURE=15.0, filter="R"
    )

    os.chdir(original_dir)

    return n_test, test_dir


def sample_directory_with_files():
    """
    Returns the path to the small sample directory used
    in the tests of ``ImageFileCollection``. Primarily intended
    for use in the doctests.
    """

    n_test, tmpdir = directory_for_testing()
    return tmpdir



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import astropy
import astropy.units as u
import numpy as np
import pytest
import skimage
from astropy.io import fits
from astropy.modeling import models
from astropy.nddata import CCDData, StdDevUncertainty
from astropy.units.quantity import Quantity
from astropy.utils.exceptions import AstropyUserWarning
from astropy.wcs import WCS

from ccdproc.core import (
    Keyword,
    ccd_process,
    cosmicray_lacosmic,
    cosmicray_median,
    create_deviation,
    flat_correct,
    gain_correct,
    subtract_bias,
    subtract_dark,
    subtract_overscan,
    transform_image,
    trim_image,
    wcs_project,
)
from ccdproc.tests.pytest_fixtures import ccd_data as ccd_data_func

try:
    from ..core import block_average, block_reduce, block_replicate

    HAS_BLOCK_X_FUNCS = True
except ImportError:
    HAS_BLOCK_X_FUNCS = False

_NUMPY_COPY_IF_NEEDED = False if np.__version__.startswith("1.") else None


# Test creating deviation
# Success expected if u_image * u_gain = u_readnoise
@pytest.mark.parametrize(
    "u_image,u_gain,u_readnoise,expect_success",
    [
        (u.electron, None, u.electron, True),
        (u.electron, u.electron, u.electron, False),
        (u.adu, u.electron / u.adu, u.electron, True),
        (u.electron, None, u.dimensionless_unscaled, False),
        (u.electron, u.dimensionless_unscaled, u.electron, True),
        (u.adu, u.dimensionless_unscaled, u.electron, False),
        (u.adu, u.photon / u.adu, u.electron, False),
    ],
)
def test_create_deviation(u_image, u_gain, u_readnoise, expect_success):
    ccd_data = ccd_data_func(data_size=10, data_mean=100)
    ccd_data.unit = u_image
    if u_gain is not None:
        gain = 2.0 * u_gain
    else:
        gain = None
    readnoise = 5 * u_readnoise
    if expect_success:
        ccd_var = create_deviation(ccd_data, gain=gain, readnoise=readnoise)
        assert ccd_var.uncertainty.array.shape == (10, 10)
        assert ccd_var.uncertainty.array.size == 100
        assert ccd_var.uncertainty.array.dtype == np.dtype(float)
        if gain is not None:
            expected_var = np.sqrt(2 * ccd_data.data + 5**2) / 2
        else:
            expected_var = np.sqrt(ccd_data.data + 5**2)
        np.testing.assert_allclose(ccd_var.uncertainty.array, expected_var)
        assert ccd_var.unit == ccd_data.unit
        # Uncertainty should *not* have any units -- does it?
        with pytest.raises(AttributeError):
            ccd_var.uncertainty.array.unit  # noqa B018 useless expression
    else:
        with pytest.raises(u.UnitsError):
            ccd_var = create_deviation(ccd_data, gain=gain, readnoise=readnoise)


def test_create_deviation_from_negative():
    ccd_data = ccd_data_func(data_mean=0, data_scale=10)
    ccd_data.unit = u.electron
    readnoise = 5 * u.electron
    ccd_var = create_deviation(
        ccd_data, gain=None, readnoise=readnoise, disregard_nan=False
    )
    np.testing.assert_array_equal(
        ccd_data.data < 0, np.isnan(ccd_var.uncertainty.array)
    )


def test_create_deviation_from_negative_2():
    ccd_data = ccd_data_func(data_mean=0, data_scale=10)
    ccd_data.unit = u.electron
    readnoise = 5 * u.electron
    ccd_var = create_deviation(
        ccd_data, gain=None, readnoise=readnoise, disregard_nan=True
    )
    mask = ccd_data.data < 0
    ccd_data.data[mask] = 0
    expected_var = np.sqrt(ccd_data.data + readnoise.value**2)
    np.testing.assert_allclose(ccd_var.uncertainty.array, expected_var)


def test_create_deviation_keywords_must_have_unit():
    ccd_data = ccd_data_func()
    # Gain must have units if provided
    with pytest.raises(TypeError):
        create_deviation(ccd_data, gain=3)
    # Readnoise must have units
    with pytest.raises(TypeError):
        create_deviation(ccd_data, readnoise=5)
    # Readnoise must be provided
    with pytest.raises(ValueError):
        create_deviation(ccd_data)


# Tests for overscan
@pytest.mark.parametrize("data_rectangle", [False, True])
@pytest.mark.parametrize(
    "median,transpose",
    [
        (False, False),
        (False, True),
        (True, False),
    ],
)
def test_subtract_overscan(median, transpose, data_rectangle):
    ccd_data = ccd_data_func()
    # Make data non-square if desired
    if data_rectangle:
        ccd_data.data = ccd_data.data[:, :-30]

    # Create the overscan region
    oscan = 300.0
    oscan_region = (slice(None), slice(0, 10))  # Indices 0 through 9
    fits_section = "[1:10, :]"
    science_region = (slice(None), slice(10, None))

    overscan_axis = 1
    if transpose:
        # Put overscan in first axis, not second, a test for #70
        oscan_region = oscan_region[::-1]
        fits_section = "[:, 1:10]"
        science_region = science_region[::-1]
        overscan_axis = 0

    ccd_data.data[oscan_region] = oscan
    # Add a fake sky background so the "science" part of the image has a
    # different average than the "overscan" part.
    sky = 10.0
    original_mean = ccd_data.data[science_region].mean()
    ccd_data.data[science_region] += oscan + sky
    # Test once using the overscan argument to specify the overscan region
    ccd_data_overscan = subtract_overscan(
        ccd_data,
        overscan=ccd_data[oscan_region],
        overscan_axis=overscan_axis,
        median=median,
        model=None,
    )
    # Is the mean of the "science" region the sum of sky and the mean the
    # "science" section had before backgrounds were added?
    np.testing.assert_almost_equal(
        ccd_data_overscan.data[science_region].mean(), sky + original_mean
    )
    # Is the overscan region zero?
    assert (ccd_data_overscan.data[oscan_region] == 0).all()

    # Now do what should be the same subtraction, with the overscan specified
    # with the fits_section
    ccd_data_fits_section = subtract_overscan(
        ccd_data,
        overscan_axis=overscan_axis,
        fits_section=fits_section,
        median=median,
        model=None,
    )
    # Is the mean of the "science" region the sum of sky and the mean the
    # "science" section had before backgrounds were added?
    np.testing.assert_almost_equal(
        ccd_data_fits_section.data[science_region].mean(), sky + original_mean
    )
    # Is the overscan region zero?
    assert (ccd_data_fits_section.data[oscan_region] == 0).all()

    # Do both ways of subtracting overscan give exactly the same result?
    np.testing.assert_allclose(
        ccd_data_overscan[science_region], ccd_data_fits_section[science_region]
    )

    # Set overscan_axis to None, and let the routine figure out the axis.
    # This should lead to the same results as before.
    ccd_data_overscan_auto = subtract_overscan(
        ccd_data,
        overscan_axis=None,
        overscan=ccd_data[oscan_region],
        median=median,
        model=None,
    )
    np.testing.assert_almost_equal(
        ccd_data_overscan_auto.data[science_region].mean(), sky + original_mean
    )
    # Use overscan_axis=None with a FITS section
    ccd_data_fits_section_overscan_auto = subtract_overscan(
        ccd_data,
        overscan_axis=None,
        fits_section=fits_section,
        median=median,
        model=None,
    )
    np.testing.assert_almost_equal(
        ccd_data_fits_section_overscan_auto.data[science_region].mean(),
        sky + original_mean,
    )
    # Overscan_axis should be 1 for a square overscan region
    # This test only works for a non-square data region, but the
    # default has the wrong axis.
    if data_rectangle:
        ccd_data.data = ccd_data.data.T
        oscan_region = (slice(None), slice(0, -30))
        science_region = (slice(None), slice(-30, None))
        ccd_data_square_overscan_auto = subtract_overscan(
            ccd_data,
            overscan_axis=None,
            overscan=ccd_data[oscan_region],
            median=median,
            model=None,
        )
        ccd_data_square = subtract_overscan(
            ccd_data,
            overscan_axis=1,
            overscan=ccd_data[oscan_region],
            median=median,
            model=None,
        )
        np.testing.assert_allclose(ccd_data_square_overscan_auto, ccd_data_square)


# A more substantial test of overscan modeling
@pytest.mark.parametrize("transpose", [True, False])
def test_subtract_overscan_model(transpose):
    ccd_data = ccd_data_func()
    # Create the overscan region
    size = ccd_data.shape[0]

    oscan_region = (slice(None), slice(0, 10))
    science_region = (slice(None), slice(10, None))

    yscan, xscan = np.mgrid[0:size, 0:size] / 10.0 + 300.0

    if transpose:
        oscan_region = oscan_region[::-1]
        science_region = science_region[::-1]
        scan = xscan
        overscan_axis = 0
    else:
        overscan_axis = 1
        scan = yscan

    original_mean = ccd_data.data[science_region].mean()

    ccd_data.data[oscan_region] = 0.0  # Only want overscan in that region
    ccd_data.data = ccd_data.data + scan

    ccd_data = subtract_overscan(
        ccd_data,
        overscan=ccd_data[oscan_region],
        overscan_axis=overscan_axis,
        median=False,
        model=models.Polynomial1D(2),
    )
    np.testing.assert_almost_equal(ccd_data.data[science_region].mean(), original_mean)
    # Set the overscan_axis explicitly to None, and let the routine
    # figure it out.
    ccd_data = subtract_overscan(
        ccd_data,
        overscan=ccd_data[oscan_region],
        overscan_axis=None,
        median=False,
        model=models.Polynomial1D(2),
    )
    np.testing.assert_almost_equal(ccd_data.data[science_region].mean(), original_mean)


def test_subtract_overscan_fails():
    ccd_data = ccd_data_func()
    # Do we get an error if the *image* is neither CCDData nor an array?
    with pytest.raises(TypeError):
        subtract_overscan(3, np.zeros((5, 5)))
    # Do we get an error if the *overscan* is not an image or an array?
    with pytest.raises(TypeError):
        subtract_overscan(np.zeros((10, 10)), 3, median=False, model=None)
    # Do we get an error if we specify both overscan and fits_section?
    with pytest.raises(TypeError):
        subtract_overscan(ccd_data, overscan=ccd_data[0:10], fits_section="[1:10]")
    # Do we raise an error if we specify neither overscan nor fits_section?
    with pytest.raises(TypeError):
        subtract_overscan(ccd_data)
    # Does a fits_section which is not a string raise an error?
    with pytest.raises(TypeError):
        subtract_overscan(ccd_data, fits_section=5)
    # Do we raise an error if the input is a plain array?
    with pytest.raises(TypeError):
        subtract_overscan(np.zeros((10, 10)), fits_section="[1:10]")


def test_trim_image_fits_section_requires_string():
    ccd_data = ccd_data_func()
    with pytest.raises(TypeError):
        trim_image(ccd_data, fits_section=5)


@pytest.mark.parametrize("mask_data, uncertainty", [(False, False), (True, True)])
def test_trim_image_fits_section(mask_data, uncertainty):
    ccd_data = ccd_data_func(data_size=50)
    if mask_data:
        ccd_data.mask = np.zeros_like(ccd_data)
    if uncertainty:
        err = np.random.default_rng().normal(size=ccd_data.shape)
        ccd_data.uncertainty = StdDevUncertainty(err)

    trimmed = trim_image(ccd_data, fits_section="[20:40,:]")
    # FITS reverse order, bounds are inclusive and starting index is 1-based
    assert trimmed.shape == (50, 21)
    np.testing.assert_allclose(trimmed.data, ccd_data[:, 19:40])
    if mask_data:
        assert trimmed.shape == trimmed.mask.shape
    if uncertainty:
        assert trimmed.shape == trimmed.uncertainty.array.shape


def test_trim_image_no_section():
    ccd_data = ccd_data_func(data_size=50)
    trimmed = trim_image(ccd_data[:, 19:40])
    assert trimmed.shape == (50, 21)
    np.testing.assert_allclose(trimmed.data, ccd_data[:, 19:40])


def test_trim_with_wcs_alters_wcs():
    ccd_data = ccd_data_func()
    # WCS construction example pulled form astropy.wcs docs
    wcs = WCS(naxis=2)
    wcs.wcs.crpix = np.array(ccd_data.shape) / 2
    wcs.wcs.cdelt = np.array([-0.066667, 0.066667])
    wcs.wcs.crval = [0, -90]
    wcs.wcs.ctype = ["RA---AIR", "DEC--AIR"]
    wcs.wcs.set_pv([(2, 1, 45.0)])
    ccd_wcs = CCDData(ccd_data, wcs=wcs)
    # The trim below should subtract 10 from the 2nd element of crpix.
    # (Second element because the FITS convention for index ordering is
    #  opposite that of python)
    trimmed = trim_image(ccd_wcs[10:, :])
    assert trimmed.wcs.wcs.crpix[1] == wcs.wcs.crpix[1] - 10


def test_subtract_bias():
    ccd_data = ccd_data_func()
    data_avg = ccd_data.data.mean()
    bias_level = 5.0
    ccd_data.data = ccd_data.data + bias_level
    ccd_data.header["key"] = "value"
    master_bias_array = np.zeros_like(ccd_data.data) + bias_level
    master_bias = CCDData(master_bias_array, unit=ccd_data.unit)
    no_bias = subtract_bias(ccd_data, master_bias, add_keyword=None)
    # Does the data we are left with have the correct average?
    np.testing.assert_almost_equal(no_bias.data.mean(), data_avg)
    # With logging turned off, metadata should not change
    assert no_bias.header == ccd_data.header
    del no_bias.header["key"]
    assert "key" in ccd_data.header
    assert no_bias.header is not ccd_data.header


def test_subtract_bias_fails():
    ccd_data = ccd_data_func(data_size=50)
    # Should fail if shapes don't match
    bias = CCDData(np.array([200, 200]), unit=u.adu)
    with pytest.raises(ValueError):
        subtract_bias(ccd_data, bias)
    # Should fail because units don't match
    bias = CCDData(np.zeros_like(ccd_data), unit=u.meter)
    with pytest.raises(u.UnitsError):
        subtract_bias(ccd_data, bias)


@pytest.mark.parametrize("exposure_keyword", [True, False])
@pytest.mark.parametrize("explicit_times", [True, False])
@pytest.mark.parametrize("scale", [True, False])
def test_subtract_dark(explicit_times, scale, exposure_keyword):
    ccd_data = ccd_data_func()
    exptime = 30.0
    exptime_key = "exposure"
    exposure_unit = u.second
    dark_level = 1.7
    master_dark_data = np.zeros_like(ccd_data.data) + dark_level
    master_dark = CCDData(master_dark_data, unit=u.adu)
    master_dark.header[exptime_key] = 2 * exptime
    dark_exptime = master_dark.header[exptime_key]
    ccd_data.header[exptime_key] = exptime
    dark_exposure_unit = exposure_unit
    if explicit_times:
        # Test case when units of dark and data exposures are different
        dark_exposure_unit = u.minute
        dark_sub = subtract_dark(
            ccd_data,
            master_dark,
            dark_exposure=dark_exptime * dark_exposure_unit,
            data_exposure=exptime * exposure_unit,
            scale=scale,
            add_keyword=None,
        )
    elif exposure_keyword:
        key = Keyword(exptime_key, unit=u.second)
        dark_sub = subtract_dark(
            ccd_data, master_dark, exposure_time=key, scale=scale, add_keyword=None
        )
    else:
        dark_sub = subtract_dark(
            ccd_data,
            master_dark,
            exposure_time=exptime_key,
            exposure_unit=u.second,
            scale=scale,
            add_keyword=None,
        )

    dark_scale = 1.0
    if scale:
        dark_scale = float(
            (exptime / dark_exptime) * (exposure_unit / dark_exposure_unit)
        )

    np.testing.assert_allclose(ccd_data.data - dark_scale * dark_level, dark_sub.data)
    # Headers should have the same content...do they?
    assert dark_sub.header == ccd_data.header
    # But the headers should not be the same object -- a copy was made
    assert dark_sub.header is not ccd_data.header


def test_subtract_dark_fails():
    ccd_data = ccd_data_func()
    # None of these tests check a result so the content of the master
    # can be anything.
    ccd_data.header["exptime"] = 30.0
    master = ccd_data.copy()

    # Do we fail if we give one of dark_exposure, data_exposure but not both?
    with pytest.raises(TypeError):
        subtract_dark(ccd_data, master, dark_exposure=30 * u.second)
    with pytest.raises(TypeError):
        subtract_dark(ccd_data, master, data_exposure=30 * u.second)

    # Do we fail if we supply dark_exposure and data_exposure and exposure_time
    with pytest.raises(TypeError):
        subtract_dark(
            ccd_data,
            master,
            dark_exposure=10 * u.second,
            data_exposure=10 * u.second,
            exposure_time="exptime",
        )

    # Fail if we supply none of the exposure-related arguments?
    with pytest.raises(TypeError):
        subtract_dark(ccd_data, master)

    # Fail if we supply exposure time but not a unit?
    with pytest.raises(TypeError):
        subtract_dark(ccd_data, master, exposure_time="exptime")

    # Fail if ccd_data or master are not CCDData objects?
    with pytest.raises(TypeError):
        subtract_dark(ccd_data.data, master, exposure_time="exptime")
    with pytest.raises(TypeError):
        subtract_dark(ccd_data, master.data, exposure_time="exptime")

    # Fail if units do not match...

    # ...when there is no scaling?
    master = CCDData(ccd_data)
    master.unit = u.meter

    with pytest.raises(u.UnitsError) as e:
        subtract_dark(ccd_data, master, exposure_time="exptime", exposure_unit=u.second)
    assert "uncalibrated image" in str(e.value)

    # Fail when the arrays are not the same size
    with pytest.raises(ValueError):
        small_master = CCDData(ccd_data)
        small_master.data = np.zeros((1, 1))
        subtract_dark(ccd_data, small_master)


def test_unit_mismatch_behaves_as_expected():
    ccd_data = ccd_data_func()
    """
    Test to alert us to any changes in how errors are raised in astropy when units
    do not match.
    """
    bad_unit = ccd_data.copy()
    bad_unit.unit = u.meter

    if astropy.__version__.startswith("1.0"):
        expected_error = ValueError
        expected_message = "operand units"
    else:
        expected_error = u.UnitConversionError
        # Make this an empty string, which always matches. In this case
        # we are really only checking by the type of error raised.
        expected_message = ""

    # Did we raise the right error?
    with pytest.raises(expected_error) as e:
        ccd_data.subtract(bad_unit)

    # Was the error message as expected?
    assert expected_message in str(e.value)


# Test for flat correction
def test_flat_correct():
    ccd_data = ccd_data_func(data_scale=10)
    # Add metadata to header for a test below...
    ccd_data.header["my_key"] = 42
    size = ccd_data.shape[0]
    # create the flat, with some scatter
    data = 2 * np.random.default_rng().normal(loc=1.0, scale=0.05, size=(size, size))
    flat = CCDData(data, meta=fits.header.Header(), unit=ccd_data.unit)
    flat_data = flat_correct(ccd_data, flat, add_keyword=None)

    # Check that the flat was normalized
    # Should be the case that flat * flat_data = ccd_data * flat.data.mean
    # if the normalization was done correctly.
    np.testing.assert_almost_equal(
        (flat_data.data * flat.data).mean(), ccd_data.data.mean() * flat.data.mean()
    )
    np.testing.assert_allclose(
        ccd_data.data / flat_data.data, flat.data / flat.data.mean()
    )

    # Check that metadata is unchanged (since logging is turned off)
    assert flat_data.header == ccd_data.header


# Test for flat correction with min_value
def test_flat_correct_min_value():
    ccd_data = ccd_data_func()
    size = ccd_data.shape[0]

    # Create the flat
    data = 2 * np.random.default_rng().normal(loc=1.0, scale=0.05, size=(size, size))
    flat = CCDData(data, meta=fits.header.Header(), unit=ccd_data.unit)
    flat_orig_data = flat.data.copy()
    min_value = 2.1  # Should replace some, but not all, values
    flat_corrected_data = flat_correct(ccd_data, flat, min_value=min_value)
    flat_with_min = flat.copy()
    flat_with_min.data[flat_with_min.data < min_value] = min_value

    # Check that the flat was normalized. The asserts below, which look a
    # little odd, are correctly testing that
    #    flat_corrected_data = ccd_data / (flat_with_min / mean(flat_with_min))
    np.testing.assert_almost_equal(
        (flat_corrected_data.data * flat_with_min.data).mean(),
        (ccd_data.data * flat_with_min.data.mean()).mean(),
    )
    np.testing.assert_allclose(
        ccd_data.data / flat_corrected_data.data,
        flat_with_min.data / flat_with_min.data.mean(),
    )

    # Test that flat is not modified.
    assert (flat_orig_data == flat.data).all()
    assert flat_orig_data is not flat.data


def test_flat_correct_norm_value():
    ccd_data = ccd_data_func(data_scale=10)
    # Test flat correction with mean value that is different than
    # the mean of the flat frame.

    # Create the flat, with some scatter
    # Note that mean value of flat is set below and is different than
    # the mean of the flat data.
    flat_mean = 5.0
    data = np.random.default_rng().normal(loc=1.0, scale=0.05, size=ccd_data.shape)
    flat = CCDData(data, meta=fits.Header(), unit=ccd_data.unit)
    flat_data = flat_correct(ccd_data, flat, add_keyword=None, norm_value=flat_mean)

    # Check that the flat was normalized
    # Should be the case that flat * flat_data = ccd_data * flat_mean
    # if the normalization was done correctly.
    np.testing.assert_almost_equal(
        (flat_data.data * flat.data).mean(), ccd_data.data.mean() * flat_mean
    )
    np.testing.assert_allclose(ccd_data.data / flat_data.data, flat.data / flat_mean)


def test_flat_correct_norm_value_bad_value():
    ccd_data = ccd_data_func()
    # Test that flat_correct raises the appropriate error if
    # it is given a bad norm_value. Bad means <=0.

    # Create the flat, with some scatter
    data = np.random.default_rng().normal(loc=1.0, scale=0.05, size=ccd_data.shape)
    flat = CCDData(data, meta=fits.Header(), unit=ccd_data.unit)
    with pytest.raises(ValueError) as e:
        flat_correct(ccd_data, flat, add_keyword=None, norm_value=-7)
    assert "norm_value must be" in str(e.value)


# Test for deviation and for flat correction
def test_flat_correct_deviation():
    ccd_data = ccd_data_func(data_scale=10, data_mean=300)
    size = ccd_data.shape[0]
    ccd_data.unit = u.electron
    ccd_data = create_deviation(ccd_data, readnoise=5 * u.electron)
    # Create the flat
    data = 2 * np.ones((size, size))
    flat = CCDData(data, meta=fits.header.Header(), unit=ccd_data.unit)
    flat = create_deviation(flat, readnoise=0.5 * u.electron)
    ccd_data = flat_correct(ccd_data, flat)


# Test the uncertainty on the data after flat correction
def test_flat_correct_data_uncertainty():
    # Regression test for #345
    dat = CCDData(np.ones([100, 100]), unit="adu", uncertainty=np.ones([100, 100]))
    # Note flat is set to 10, error, if present, is set to one.
    flat = CCDData(10 * np.ones([100, 100]), unit="adu")
    res = flat_correct(dat, flat)
    assert (res.data == dat.data).all()
    assert (res.uncertainty.array == dat.uncertainty.array).all()


# Tests for gain correction
def test_gain_correct():
    ccd_data = ccd_data_func()
    init_data = ccd_data.data
    gain_data = gain_correct(ccd_data, gain=3, add_keyword=None)
    np.testing.assert_allclose(gain_data.data, 3 * init_data)
    assert ccd_data.meta == gain_data.meta


def test_gain_correct_quantity():
    ccd_data = ccd_data_func()
    init_data = ccd_data.data
    g = Quantity(3, u.electron / u.adu)
    ccd_data = gain_correct(ccd_data, gain=g)

    np.testing.assert_allclose(ccd_data.data, 3 * init_data)
    assert ccd_data.unit == u.electron


# Test transform is ccd
def test_transform_isccd():
    with pytest.raises(TypeError):
        transform_image(1, 1)


# Test function is callable
def test_transform_isfunc():
    ccd_data = ccd_data_func()
    with pytest.raises(TypeError):
        transform_image(ccd_data, 1)


# Test warning is issue if WCS information is available
def test_catch_transform_wcs_warning():
    ccd_data = ccd_data_func()

    def tran(arr):
        return 10 * arr

    # No warning.
    transform_image(ccd_data, tran)

    # Issue warning when data has WCS.
    ccd_data.wcs = wcs_for_testing(ccd_data.shape)
    with pytest.warns(UserWarning, match="WCS information may be incorrect"):
        transform_image(ccd_data, tran)


@pytest.mark.parametrize("mask_data, uncertainty", [(False, False), (True, True)])
def test_transform_image(mask_data, uncertainty):
    ccd_data = ccd_data_func(data_size=50)
    if mask_data:
        ccd_data.mask = np.zeros_like(ccd_data)
        ccd_data.mask[10, 10] = 1
    if uncertainty:
        err = np.random.default_rng().normal(size=ccd_data.shape)
        ccd_data.uncertainty = StdDevUncertainty(err)

    def tran(arr):
        return 10 * arr

    tran = transform_image(ccd_data, tran)

    np.testing.assert_allclose(10 * ccd_data.data, tran.data)
    if mask_data:
        assert tran.shape == tran.mask.shape
        np.testing.assert_array_equal(ccd_data.mask, tran.mask)
    if uncertainty:
        assert tran.shape == tran.uncertainty.array.shape
        np.testing.assert_allclose(
            10 * ccd_data.uncertainty.array, tran.uncertainty.array
        )


# Test block_reduce and block_replicate wrapper
@pytest.mark.skipif(not HAS_BLOCK_X_FUNCS, reason="needs astropy >= 1.1.x")
@pytest.mark.skipif(
    (skimage.__version__ < "0.14.2") and ("dev" in np.__version__),
    reason="Incompatibility between scikit-image " "and numpy 1.16",
)
def test_block_reduce():
    ccd = CCDData(
        np.ones((4, 4)),
        unit="adu",
        meta={"testkw": 1},
        mask=np.zeros((4, 4), dtype=bool),
        uncertainty=StdDevUncertainty(np.ones((4, 4))),
    )
    with pytest.warns(AstropyUserWarning) as w:
        ccd_summed = block_reduce(ccd, (2, 2))
    assert len(w) == 1
    assert "following attributes were set" in str(w[0].message)
    assert isinstance(ccd_summed, CCDData)
    assert np.all(ccd_summed.data == 4)
    assert ccd_summed.data.shape == (2, 2)
    assert ccd_summed.unit == u.adu
    # Other attributes are set to None. In case the function is modified to
    # work on these attributes correctly those tests need to be updated!
    assert ccd_summed.meta == {"testkw": 1}
    assert ccd_summed.mask is None
    assert ccd_summed.uncertainty is None

    # Make sure meta is copied
    ccd_summed.meta["testkw2"] = 10
    assert "testkw2" not in ccd.meta


@pytest.mark.skipif(not HAS_BLOCK_X_FUNCS, reason="needs astropy >= 1.1.x")
@pytest.mark.skipif(
    (skimage.__version__ < "0.14.2") and ("dev" in np.__version__),
    reason="Incompatibility between scikit-image " "and numpy 1.16",
)
def test_block_average():
    ccd = CCDData(
        np.ones((4, 4)),
        unit="adu",
        meta={"testkw": 1},
        mask=np.zeros((4, 4), dtype=bool),
        uncertainty=StdDevUncertainty(np.ones((4, 4))),
    )
    ccd.data[::2, ::2] = 2
    with pytest.warns(AstropyUserWarning) as w:
        ccd_avgd = block_average(ccd, (2, 2))
    assert len(w) == 1
    assert "following attributes were set" in str(w[0].message)

    assert isinstance(ccd_avgd, CCDData)
    assert np.all(ccd_avgd.data == 1.25)
    assert ccd_avgd.data.shape == (2, 2)
    assert ccd_avgd.unit == u.adu
    # Other attributes are set to None. In case the function is modified to
    # work on these attributes correctly those tests need to be updated!
    assert ccd_avgd.meta == {"testkw": 1}
    assert ccd_avgd.mask is None
    assert ccd_avgd.wcs is None
    assert ccd_avgd.uncertainty is None

    # Make sure meta is copied
    ccd_avgd.meta["testkw2"] = 10
    assert "testkw2" not in ccd.meta


@pytest.mark.skipif(not HAS_BLOCK_X_FUNCS, reason="needs astropy >= 1.1.x")
def test_block_replicate():
    ccd = CCDData(
        np.ones((4, 4)),
        unit="adu",
        meta={"testkw": 1},
        mask=np.zeros((4, 4), dtype=bool),
        uncertainty=StdDevUncertainty(np.ones((4, 4))),
    )
    with pytest.warns(AstropyUserWarning) as w:
        ccd_repl = block_replicate(ccd, (2, 2))
    assert len(w) == 1
    assert "following attributes were set" in str(w[0].message)

    assert isinstance(ccd_repl, CCDData)
    assert np.all(ccd_repl.data == 0.25)
    assert ccd_repl.data.shape == (8, 8)
    assert ccd_repl.unit == u.adu
    # Other attributes are set to None. In case the function is modified to
    # work on these attributes correctly those tests need to be updated!
    assert ccd_repl.meta == {"testkw": 1}
    assert ccd_repl.mask is None
    assert ccd_repl.wcs is None
    assert ccd_repl.uncertainty is None

    # Make sure meta is copied
    ccd_repl.meta["testkw2"] = 10
    assert "testkw2" not in ccd.meta


# Test overscan changes
def test__overscan_schange():
    ccd_data = ccd_data_func()
    old_data = ccd_data.copy()
    new_data = subtract_overscan(ccd_data, overscan=ccd_data[:, 1], overscan_axis=0)
    assert not np.allclose(old_data.data, new_data.data)
    np.testing.assert_allclose(old_data.data, ccd_data.data)


def test_create_deviation_does_not_change_input():
    ccd_data = ccd_data_func()
    original = ccd_data.copy()
    _ = create_deviation(
        ccd_data, gain=5 * u.electron / u.adu, readnoise=10 * u.electron
    )
    np.testing.assert_allclose(original.data, ccd_data.data)
    assert original.unit == ccd_data.unit


def test_cosmicray_median_does_not_change_input():
    ccd_data = ccd_data_func()
    original = ccd_data.copy()
    error = np.zeros_like(ccd_data)
    with np.errstate(invalid="ignore", divide="ignore"):
        _ = cosmicray_median(
            ccd_data, error_image=error, thresh=5, mbox=11, gbox=0, rbox=0
        )
    np.testing.assert_allclose(original.data, ccd_data.data)
    assert original.unit == ccd_data.unit


def test_cosmicray_lacosmic_does_not_change_input():
    ccd_data = ccd_data_func()
    original = ccd_data.copy()
    _ = cosmicray_lacosmic(ccd_data)
    np.testing.assert_allclose(original.data, ccd_data.data)
    assert original.unit == ccd_data.unit


def test_flat_correct_does_not_change_input():
    ccd_data = ccd_data_func()
    original = ccd_data.copy()
    flat = CCDData(np.zeros_like(ccd_data), unit=ccd_data.unit)
    with np.errstate(invalid="ignore"):
        _ = flat_correct(ccd_data, flat=flat)
    np.testing.assert_allclose(original.data, ccd_data.data)
    assert original.unit == ccd_data.unit


def test_gain_correct_does_not_change_input():
    ccd_data = ccd_data_func()
    original = ccd_data.copy()
    _ = gain_correct(ccd_data, gain=1, gain_unit=ccd_data.unit)
    np.testing.assert_allclose(original.data, ccd_data.data)
    assert original.unit == ccd_data.unit


def test_subtract_bias_does_not_change_input():
    ccd_data = ccd_data_func()
    original = ccd_data.copy()
    master_frame = CCDData(np.zeros_like(ccd_data), unit=ccd_data.unit)
    _ = subtract_bias(ccd_data, master=master_frame)
    np.testing.assert_allclose(original.data, ccd_data.data)
    assert original.unit == ccd_data.unit


def test_trim_image_does_not_change_input():
    ccd_data = ccd_data_func()
    original = ccd_data.copy()
    _ = trim_image(ccd_data, fits_section=None)
    np.testing.assert_allclose(original.data, ccd_data.data)
    assert original.unit == ccd_data.unit


def test_transform_image_does_not_change_input():
    ccd_data = ccd_data_func()
    original = ccd_data.copy()
    with np.errstate(invalid="ignore"):
        _ = transform_image(ccd_data, np.sqrt)
    np.testing.assert_allclose(original.data, ccd_data)
    assert original.unit == ccd_data.unit


def wcs_for_testing(shape):
    # Set up a simple WCS, details are cut/pasted from astropy WCS docs,
    # mostly. CRPIX is set to the center of shape, rounded down.

    # Create a new WCS object. The number of axes must be set
    # from the start
    w = WCS(naxis=2)

    # Set up an "Airy's zenithal" projection
    # Vector properties may be set with Python lists, or Numpy arrays
    w.wcs.crpix = [shape[0] // 2, shape[1] // 2]
    w.wcs.cdelt = np.array([-0.066667, 0.066667])
    w.wcs.crval = [0, -90]
    w.wcs.ctype = ["RA---AIR", "DEC--AIR"]
    w.wcs.set_pv([(2, 1, 45.0)])

    return w


def test_wcs_project_onto_same_wcs():
    ccd_data = ccd_data_func()
    # The trivial case, same WCS, no mask.
    target_wcs = wcs_for_testing(ccd_data.shape)
    ccd_data.wcs = wcs_for_testing(ccd_data.shape)

    new_ccd = wcs_project(ccd_data, target_wcs)

    # Make sure new image has correct WCS.
    assert new_ccd.wcs.wcs.compare(target_wcs.wcs)

    # Make sure data matches within some reasonable tolerance.
    np.testing.assert_allclose(ccd_data.data, new_ccd.data, rtol=1e-5)


def test_wcs_project_onto_same_wcs_remove_headers():
    ccd_data = ccd_data_func()
    # Remove an example WCS keyword from the header
    target_wcs = wcs_for_testing(ccd_data.shape)
    ccd_data.wcs = wcs_for_testing(ccd_data.shape)
    ccd_data.header = ccd_data.wcs.to_header()

    new_ccd = wcs_project(ccd_data, target_wcs)

    for k in ccd_data.wcs.to_header():
        assert k not in new_ccd.header


def test_wcs_project_onto_shifted_wcs():
    ccd_data = ccd_data_func()
    # Just make the target WCS the same as the initial with the center
    # pixel shifted by 1 in x and y.

    ccd_data.wcs = wcs_for_testing(ccd_data.shape)
    target_wcs = wcs_for_testing(ccd_data.shape)
    target_wcs.wcs.crpix += [1, 1]

    ccd_data.mask = np.random.default_rng().choice([0, 1], size=ccd_data.shape)

    new_ccd = wcs_project(ccd_data, target_wcs)

    # Make sure new image has correct WCS.
    assert new_ccd.wcs.wcs.compare(target_wcs.wcs)

    # Make sure data matches within some reasonable tolerance, keeping in mind
    # that the pixels should all be shifted.
    masked_input = np.ma.array(ccd_data.data, mask=ccd_data.mask)
    masked_output = np.ma.array(new_ccd.data, mask=new_ccd.mask)
    np.testing.assert_allclose(masked_input[:-1, :-1], masked_output[1:, 1:], rtol=1e-5)

    # The masks should all be shifted too.
    np.testing.assert_array_equal(ccd_data.mask[:-1, :-1], new_ccd.mask[1:, 1:])

    # We should have more values that are masked in the output array
    # than on input because some on output were not in the footprint
    # of the original array.

    # In the case of a shift, one row and one column should be nan, and they
    # will share one common nan where they intersect, so we know how many nan
    # there should be.
    assert np.isnan(new_ccd.data).sum() == np.sum(new_ccd.shape) - 1


# Use an odd number of pixels to make a well-defined center pixel
def test_wcs_project_onto_scale_wcs():
    # Make the target WCS with half the pixel scale and number of pixels
    # and the values should drop by a factor of 4.
    ccd_data = ccd_data_func(data_size=31)

    ccd_data.wcs = wcs_for_testing(ccd_data.shape)

    # Make sure wcs is centered at the center of the center pixel.
    ccd_data.wcs.wcs.crpix += 0.5

    # Use uniform input data value for simplicity.
    ccd_data.data = np.ones_like(ccd_data.data)

    # Make mask zero...
    ccd_data.mask = np.zeros_like(ccd_data.data)
    # ...except the center pixel, which is one.
    ccd_data.mask[int(ccd_data.wcs.wcs.crpix[0]), int(ccd_data.wcs.wcs.crpix[1])] = 1

    target_wcs = wcs_for_testing(ccd_data.shape)
    target_wcs.wcs.cdelt /= 2

    # Choice below ensures we are really at the center pixel of an odd range.
    target_shape = 2 * np.array(ccd_data.shape) + 1
    target_wcs.wcs.crpix = 2 * target_wcs.wcs.crpix + 1 + 0.5

    # Explicitly set the interpolation method so we know what to
    # expect for the mass.
    new_ccd = wcs_project(
        ccd_data, target_wcs, target_shape=target_shape, order="nearest-neighbor"
    )

    # Make sure new image has correct WCS.
    assert new_ccd.wcs.wcs.compare(target_wcs.wcs)

    # Define a cutout from the new array that should match the old.
    new_lower_bound = (np.array(new_ccd.shape) - np.array(ccd_data.shape)) // 2
    new_upper_bound = (np.array(new_ccd.shape) + np.array(ccd_data.shape)) // 2
    data_cutout = new_ccd.data[
        new_lower_bound[0] : new_upper_bound[0], new_lower_bound[1] : new_upper_bound[1]
    ]

    # Make sure data matches within some reasonable tolerance, keeping in mind
    # that the pixels have been scaled.
    np.testing.assert_allclose(ccd_data.data / 4, data_cutout, rtol=1e-5)

    # Mask should be true for four pixels (all nearest neighbors)
    # of the single pixel we masked initially.
    new_center = np.array(new_ccd.wcs.wcs.crpix, dtype=int, copy=_NUMPY_COPY_IF_NEEDED)
    assert np.all(
        new_ccd.mask[
            new_center[0] : new_center[0] + 2, new_center[1] : new_center[1] + 2
        ]
    )

    # Those four, and any that reproject made nan because they draw on
    # pixels outside the footprint of the original image, are the only
    # pixels that should be masked.
    assert new_ccd.mask.sum() == 4 + np.isnan(new_ccd.data).sum()


def test_ccd_process_does_not_change_input():
    ccd_data = ccd_data_func()
    original = ccd_data.copy()
    _ = ccd_process(ccd_data, gain=5 * u.electron / u.adu, readnoise=10 * u.electron)
    np.testing.assert_allclose(original.data, ccd_data.data)
    assert original.unit == ccd_data.unit


def test_ccd_process_parameters_are_appropriate():
    ccd_data = ccd_data_func()
    # oscan check
    with pytest.raises(TypeError):
        ccd_process(ccd_data, oscan=True)

    # Trim section check
    with pytest.raises(TypeError):
        ccd_process(ccd_data, trim=True)

    # Error frame check
    # gain and readnoise must be specified
    with pytest.raises(ValueError):
        ccd_process(ccd_data, error=True)

    # gain must be specified
    with pytest.raises(ValueError):
        ccd_process(ccd_data, error=True, gain=None, readnoise=5)

    # Mask check
    with pytest.raises(TypeError):
        ccd_process(ccd_data, bad_pixel_mask=3)

    # master bias check
    with pytest.raises(TypeError):
        ccd_process(ccd_data, master_bias=3)

    # Master flat check
    with pytest.raises(TypeError):
        ccd_process(ccd_data, master_flat=3)


def test_ccd_process():
    # Test the through ccd_process
    ccd_data = CCDData(10.0 * np.ones((100, 100)), unit=u.adu)
    ccd_data.data[:, -10:] = 2
    ccd_data.meta["testkw"] = 100

    mask = np.zeros((100, 90))

    masterbias = CCDData(2.0 * np.ones((100, 90)), unit=u.electron)
    masterbias.uncertainty = StdDevUncertainty(np.zeros((100, 90)))

    dark_frame = CCDData(0.0 * np.ones((100, 90)), unit=u.electron)
    dark_frame.uncertainty = StdDevUncertainty(np.zeros((100, 90)))

    masterflat = CCDData(10.0 * np.ones((100, 90)), unit=u.electron)
    masterflat.uncertainty = StdDevUncertainty(np.zeros((100, 90)))

    occd = ccd_process(
        ccd_data,
        oscan=ccd_data[:, -10:],
        trim="[1:90,1:100]",
        error=True,
        master_bias=masterbias,
        master_flat=masterflat,
        dark_frame=dark_frame,
        bad_pixel_mask=mask,
        gain=0.5 * u.electron / u.adu,
        readnoise=5**0.5 * u.electron,
        oscan_median=True,
        dark_scale=False,
        dark_exposure=1.0 * u.s,
        data_exposure=1.0 * u.s,
    )

    # Final results should be (10 - 2) / 2.0 - 2 = 2
    # Error should be (4 + 5)**0.5 / 0.5  = 3.0

    np.testing.assert_allclose(2.0 * np.ones((100, 90)), occd.data)
    np.testing.assert_almost_equal(3.0 * np.ones((100, 90)), occd.uncertainty.array)
    np.testing.assert_array_equal(mask, occd.mask)
    assert occd.unit == u.electron
    # Make sure the original keyword is still present. Regression test for #401
    assert occd.meta["testkw"] == 100


def test_ccd_process_gain_corrected():
    # Test the through ccd_process with gain_corrected as False
    ccd_data = CCDData(10.0 * np.ones((100, 100)), unit=u.adu)
    ccd_data.data[:, -10:] = 2
    ccd_data.meta["testkw"] = 100

    mask = np.zeros((100, 90))

    masterbias = CCDData(4.0 * np.ones((100, 90)), unit=u.adu)
    masterbias.uncertainty = StdDevUncertainty(np.zeros((100, 90)))

    dark_frame = CCDData(0.0 * np.ones((100, 90)), unit=u.adu)
    dark_frame.uncertainty = StdDevUncertainty(np.zeros((100, 90)))

    masterflat = CCDData(5.0 * np.ones((100, 90)), unit=u.adu)
    masterflat.uncertainty = StdDevUncertainty(np.zeros((100, 90)))

    occd = ccd_process(
        ccd_data,
        oscan=ccd_data[:, -10:],
        trim="[1:90,1:100]",
        error=True,
        master_bias=masterbias,
        master_flat=masterflat,
        dark_frame=dark_frame,
        bad_pixel_mask=mask,
        gain=0.5 * u.electron / u.adu,
        readnoise=5**0.5 * u.electron,
        oscan_median=True,
        dark_scale=False,
        dark_exposure=1.0 * u.s,
        data_exposure=1.0 * u.s,
        gain_corrected=False,
    )

    # Final results should be (10 - 2) / 2.0 - 2 = 2
    # Error should be (4 + 5)**0.5 / 0.5  = 3.0

    np.testing.assert_allclose(2.0 * np.ones((100, 90)), occd.data)
    np.testing.assert_almost_equal(3.0 * np.ones((100, 90)), occd.uncertainty.array)
    np.testing.assert_array_equal(mask, occd.mask)
    assert occd.unit == u.electron
    # Make sure the original keyword is still present. Regression test for #401
    assert occd.meta["testkw"] == 100



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import logging
import os
from glob import iglob
from pathlib import Path
from shutil import rmtree
from tempfile import NamedTemporaryFile, TemporaryDirectory, mkdtemp

import astropy.io.fits as fits
import numpy as np
import pytest
from astropy.io.fits.verify import VerifyWarning
from astropy.nddata import CCDData
from astropy.table import Table
from astropy.utils.data import get_pkg_data_filename
from astropy.utils.exceptions import AstropyUserWarning

from ccdproc.image_collection import ImageFileCollection

_filters = []
_original_dir = ""


def test_fits_summary(triage_setup):
    keywords = ["imagetyp", "filter"]
    ic = ImageFileCollection(triage_setup.test_dir, keywords=keywords)
    summary = ic._fits_summary(header_keywords=keywords)
    assert len(summary["file"]) == triage_setup.n_test["files"]
    for keyword in keywords:
        assert len(summary[keyword]) == triage_setup.n_test["files"]
    # Explicit conversion to array is needed to avoid astropy Table bug in
    # 0.2.4
    no_filter_no_object_row = np.array(
        summary["file"] == "no_filter_no_object_bias.fit"
    )
    # There should be no filter keyword in the bias file
    assert summary["filter"][no_filter_no_object_row].mask


class TestImageFileCollectionRepresentation:
    def test_repr_location(self, triage_setup):
        ic = ImageFileCollection(location=triage_setup.test_dir)
        assert repr(ic) == f"ImageFileCollection(location={triage_setup.test_dir!r})"

    def test_repr_keywords(self, triage_setup):
        ic = ImageFileCollection(location=triage_setup.test_dir, keywords=["imagetyp"])
        ref = (
            f"ImageFileCollection(location={triage_setup.test_dir!r}, "
            "keywords=['imagetyp'])"
        )
        assert repr(ic) == ref

    def test_repr_globs(self, triage_setup):
        ic = ImageFileCollection(
            location=triage_setup.test_dir,
            glob_exclude="*no_filter*",
            glob_include="*object_light*",
        )
        ref = (
            f"ImageFileCollection(location={triage_setup.test_dir!r}, "
            "glob_include='*object_light*', "
            "glob_exclude='*no_filter*')"
        )
        assert repr(ic) == ref

    def test_repr_files(self, triage_setup):
        ic = ImageFileCollection(
            location=triage_setup.test_dir,
            filenames=["no_filter_no_object_light.fit", "no_filter_no_object_bias.fit"],
        )
        ref = (
            f"ImageFileCollection(location={triage_setup.test_dir!r}, "
            "filenames=['no_filter_no_object_light.fit', "
            "'no_filter_no_object_bias.fit'])"
        )
        assert repr(ic) == ref

    def test_repr_ext(self, triage_setup):

        hdul = fits.HDUList(
            [fits.PrimaryHDU(np.ones((10, 10))), fits.ImageHDU(np.ones((10, 10)))]
        )
        hdul.writeto(os.path.join(triage_setup.test_dir, "mef.fits"))

        ic = ImageFileCollection(
            location=triage_setup.test_dir, filenames=["mef.fits"], ext=1
        )
        ref = (
            f"ImageFileCollection(location={triage_setup.test_dir!r}, "
            "filenames=['mef.fits'], "
            "ext=1)"
        )
        assert repr(ic) == ref


# This should work mark all test methods as using the triage_setup
# fixture, but it doesn't, so the fixture is given explicitly as an
# argument to each method.
# @pytest.mark.usefixtures("triage_setup")
class TestImageFileCollection:
    def _basenames(self, paths):
        return set([os.path.basename(file) for file in paths])

    def _setup_logger(self, path, level=logging.WARN):
        """
        Set up file logger at the path.
        """
        logger = logging.getLogger()
        logger.setLevel(level)
        logger.addHandler(logging.FileHandler(path))
        return logger

    def test_filter_files(self, triage_setup):
        img_collection = ImageFileCollection(
            location=triage_setup.test_dir, keywords=["imagetyp", "filter"]
        )
        assert (
            len(img_collection.files_filtered(imagetyp="bias"))
            == triage_setup.n_test["bias"]
        )
        assert len(img_collection.files) == triage_setup.n_test["files"]
        assert "filter" in img_collection.keywords
        assert "flying monkeys" not in img_collection.keywords
        assert len(img_collection.values("imagetyp", unique=True)) == 2

    def test_filter_files_whitespace_keys(self, triage_setup):
        hdr = fits.Header([("HIERARCH a b", 2)])
        hdul = fits.HDUList([fits.PrimaryHDU(np.ones((10, 10)), header=hdr)])
        hdul.writeto(os.path.join(triage_setup.test_dir, "hdr_with_whitespace.fits"))

        ic = ImageFileCollection(location=triage_setup.test_dir)
        # Using a dictionary and unpacking it should work
        filtered = ic.files_filtered(**{"a b": 2})
        assert len(filtered) == 1
        assert "hdr_with_whitespace.fits" in filtered

        # Also check it's working with generators:
        for _, filename in ic.data(a_b=2, replace_="_", return_fname=True):
            assert filename == "hdr_with_whitespace.fits"

    def test_filter_files_with_str_on_nonstr_column(self, triage_setup):
        ic = ImageFileCollection(location=triage_setup.test_dir)
        # Filtering an integer column with a string
        filtered = ic.files_filtered(naxis="2")
        assert len(filtered) == 0

    def test_filter_fz_files(self, triage_setup):
        fn = "test.fits.fz"
        ic = ImageFileCollection(location=triage_setup.test_dir, filenames=fn)
        # Get a subset of files with a specific header value
        filtered = ic.files_filtered(exposure=15.0)
        assert len(filtered) == 1

    def test_filtered_files_have_proper_path(self, triage_setup):
        ic = ImageFileCollection(location=triage_setup.test_dir, keywords="*")
        # Get a subset of the files.
        plain_biases = ic.files_filtered(imagetyp="bias")
        # Force a copy...
        plain_biases = list(plain_biases)
        # Same subset, but with full path.
        path_biases = ic.files_filtered(imagetyp="bias", include_path=True)
        for path_b, plain_b in zip(path_biases, plain_biases):
            # If the path munging has been done properly, this will succeed.
            assert os.path.basename(path_b) == plain_b

    def test_filenames_are_set_properly(self, triage_setup):
        fn = ["filter_no_object_bias.fit", "filter_object_light_foo.fit"]
        img_collection = ImageFileCollection(
            location=triage_setup.test_dir, filenames=fn, keywords=["filter"]
        )
        assert img_collection.files == fn

        img_collection.refresh()
        assert img_collection.files == fn

        fn = "filter_no_object_bias.fit"
        img_collection = ImageFileCollection(
            location=triage_setup.test_dir, filenames=fn, keywords=["filter"]
        )
        assert img_collection.files == [fn]

    def test_keywords_deleter(self, triage_setup):
        ic = ImageFileCollection(triage_setup.test_dir, keywords="*")

        assert ic.keywords != []
        del ic.keywords
        assert ic.keywords == []

    def test_files_with_compressed(self, triage_setup):
        collection = ImageFileCollection(location=triage_setup.test_dir)
        assert (
            len(collection._fits_files_in_directory(compressed=True))
            == triage_setup.n_test["files"]
        )

    def test_files_with_no_compressed(self, triage_setup):
        collection = ImageFileCollection(location=triage_setup.test_dir)
        n_files_found = len(collection._fits_files_in_directory(compressed=False))
        n_uncompressed = (
            triage_setup.n_test["files"] - triage_setup.n_test["compressed"]
        )
        assert n_files_found == n_uncompressed

    def test_generator_full_path(self, triage_setup):
        collection = ImageFileCollection(
            location=triage_setup.test_dir, keywords=["imagetyp"]
        )

        for path, file_name in zip(collection._paths(), collection.files):
            assert path == os.path.join(triage_setup.test_dir, file_name)

    def test_hdus(self, triage_setup):
        collection = ImageFileCollection(
            location=triage_setup.test_dir, keywords=["imagetyp"]
        )

        n_hdus = 0
        for hdu in collection.hdus():
            assert isinstance(hdu, fits.PrimaryHDU)
            data = hdu.data  # Must access the data to force scaling
            # pre-astropy 1.1 unsigned data was changed to float32 and BZERO
            # removed. In 1.1 and later, BZERO stays but the data type is
            # unsigned int.
            assert ("BZERO" not in hdu.header) or (data.dtype is np.dtype(np.uint16))
            n_hdus += 1
        assert n_hdus == triage_setup.n_test["files"]

    def test_hdus_masking(self, triage_setup):
        collection = ImageFileCollection(
            location=triage_setup.test_dir, keywords=["imagetyp", "exposure"]
        )
        old_data = np.array(collection.summary)
        for _ in collection.hdus(imagetyp="bias"):
            pass
        new_data = np.array(collection.summary)
        assert (new_data == old_data).all()

    @pytest.mark.parametrize("extension", ["TESTEXT", 1, ("TESTEXT", 1)])
    def test_multiple_extensions(self, triage_setup, extension):
        ext1 = fits.PrimaryHDU()
        ext1.data = np.arange(1, 5)
        # It is important than the name used for this test extension
        # NOT be MASK or UNCERT because both are treated in a special
        # way by the FITS reader.
        test_ext_name = "TESTEXT"
        ext2 = fits.ImageHDU(name=test_ext_name)
        ext2.data = np.arange(6, 10)
        hdulist = fits.hdu.hdulist.HDUList([ext1, ext2])

        hdulist.writeto(os.path.join(triage_setup.test_dir, "multi-extension.fits"))
        ic2 = ImageFileCollection(
            triage_setup.test_dir,
            keywords="*",
            filenames=["multi-extension.fits"],
            ext=extension,
        )

        ic1 = ImageFileCollection(
            triage_setup.test_dir,
            keywords="*",
            filenames=["multi-extension.fits"],
            ext=0,
        )

        assert ic1.ext == 0
        assert ic2.ext == extension

        column2 = ic2.summary.colnames
        column1 = ic1.summary.colnames

        assert column1 != column2

        list1 = [key.lower() for key in ext2.header]
        list2 = ic2.summary.colnames[1:]

        assert list1 == list2

        ccd_kwargs = {"unit": "adu"}
        for data, hdr, hdu, ccd in zip(
            ic2.data(), ic2.headers(), ic2.hdus(), ic2.ccds(ccd_kwargs)
        ):
            np.testing.assert_allclose(data, ext2.data)
            assert hdr == ext2.header
            # Now compare that the generators each give the same stuff
            np.testing.assert_allclose(data, ccd.data)
            np.testing.assert_allclose(data, hdu.data)
            assert hdr == hdu.header
            assert hdr == ccd.meta

    def test_headers(self, triage_setup):
        collection = ImageFileCollection(
            location=triage_setup.test_dir, keywords=["imagetyp"]
        )
        n_headers = 0
        for header in collection.headers():
            assert isinstance(header, fits.Header)
            assert "bzero" in header
            n_headers += 1
        assert n_headers == triage_setup.n_test["files"]

    def test_headers_save_location(self, triage_setup):
        collection = ImageFileCollection(
            location=triage_setup.test_dir, keywords=["imagetyp"]
        )
        destination = mkdtemp()
        for _ in collection.headers(save_location=destination):
            pass
        new_collection = ImageFileCollection(
            location=destination, keywords=["imagetyp"]
        )

        assert (
            len(
                self._basenames(collection._paths())
                - self._basenames(new_collection._paths())
            )
            == 0
        )
        rmtree(destination)

    def test_headers_with_filter(self, triage_setup):
        collection = ImageFileCollection(
            location=triage_setup.test_dir, keywords=["imagetyp"]
        )
        cnt = 0
        for header in collection.headers(imagetyp="light"):
            assert header["imagetyp"].lower() == "light"
            cnt += 1
        assert cnt == triage_setup.n_test["light"]

    def test_headers_with_multiple_filters(self, triage_setup):
        collection = ImageFileCollection(
            location=triage_setup.test_dir, keywords=["imagetyp"]
        )
        cnt = 0
        for header in collection.headers(imagetyp="light", filter="R"):
            assert header["imagetyp"].lower() == "light"
            assert header["filter"].lower() == "r"
            cnt += 1
        assert cnt == (
            triage_setup.n_test["light"] - triage_setup.n_test["missing_filter_value"]
        )

    def test_headers_with_filter_wildcard(self, triage_setup):
        collection = ImageFileCollection(
            location=triage_setup.test_dir, keywords=["imagetyp"]
        )
        cnt = 0
        for _ in collection.headers(imagetyp="*"):
            cnt += 1
        assert cnt == triage_setup.n_test["files"]

    def test_headers_with_filter_missing_keyword(self, triage_setup):
        collection = ImageFileCollection(
            location=triage_setup.test_dir, keywords=["imagetyp"]
        )
        for header in collection.headers(imagetyp="light", object=""):
            assert header["imagetyp"].lower() == "light"
            with pytest.raises(KeyError):
                header["object"]

    def test_generator_headers_save_with_name(self, triage_setup):
        collection = ImageFileCollection(
            location=triage_setup.test_dir, keywords=["imagetyp"]
        )
        for header in collection.headers(save_with_name="_new"):
            assert isinstance(header, fits.Header)
        new_collection = ImageFileCollection(
            location=triage_setup.test_dir, keywords=["imagetyp"]
        )
        assert (
            len(new_collection._paths())
            == 2 * (triage_setup.n_test["files"]) - triage_setup.n_test["compressed"]
        )
        [os.remove(fil) for fil in iglob(triage_setup.test_dir + "/*_new*")]

    def test_generator_data(self, triage_setup):
        collection = ImageFileCollection(
            location=triage_setup.test_dir, keywords=["imagetyp"]
        )
        for img in collection.data():
            assert isinstance(img, np.ndarray)

    def test_generator_ccds_without_unit(self, triage_setup):
        collection = ImageFileCollection(
            location=triage_setup.test_dir, keywords=["imagetyp"]
        )

        with pytest.raises(ValueError):
            _ = next(collection.ccds())

    def test_generator_ccds(self, triage_setup):
        collection = ImageFileCollection(
            location=triage_setup.test_dir, keywords=["imagetyp"]
        )
        ccd_kwargs = {"unit": "adu"}
        for ccd in collection.ccds(ccd_kwargs=ccd_kwargs):
            assert isinstance(ccd, CCDData)

    def test_consecutive_fiilters(self, triage_setup):
        collection = ImageFileCollection(
            location=triage_setup.test_dir, keywords=["imagetyp", "filter", "object"]
        )
        no_files_match = collection.files_filtered(object="fdsafs")
        assert len(no_files_match) == 0
        some_files_should_match = collection.files_filtered(
            object=None, imagetyp="light"
        )
        assert len(some_files_should_match) == triage_setup.n_test["light"]

    def test_filter_does_not_not_permanently_change_file_mask(self, triage_setup):
        collection = ImageFileCollection(
            location=triage_setup.test_dir, keywords=["imagetyp"]
        )
        # Ensure all files are originally unmasked
        assert not collection.summary["file"].mask.any()
        # Generate list that will match NO files
        collection.files_filtered(imagetyp="foisajfoisaj")
        # If the code works, this should have no permanent effect
        assert not collection.summary["file"].mask.any()

    @pytest.mark.parametrize(
        "new_keywords,collection_keys",
        [
            (["imagetyp", "object"], ["imagetyp", "filter"]),
            (["imagetyp"], ["imagetyp", "filter"]),
        ],
    )
    def test_keyword_setting(self, new_keywords, collection_keys, triage_setup):
        collection = ImageFileCollection(
            location=triage_setup.test_dir, keywords=collection_keys
        )
        tbl_orig = collection.summary
        collection.keywords = new_keywords
        tbl_new = collection.summary

        if set(new_keywords).issubset(collection_keys):
            # Should just delete columns without rebuilding table
            assert tbl_orig is tbl_new
        else:
            # We need new keywords so must rebuild
            assert tbl_orig is not tbl_new

        for key in new_keywords:
            assert key in tbl_new.keys()
        assert (tbl_orig["file"] == tbl_new["file"]).all()
        assert (tbl_orig["imagetyp"] == tbl_new["imagetyp"]).all()
        assert "filter" not in tbl_new.keys()
        assert "object" not in tbl_orig.keys()

    def test_keyword_setting_to_empty_list(self, triage_setup):
        ic = ImageFileCollection(triage_setup.test_dir)
        ic.keywords = []
        assert ["file"] == ic.keywords

    def test_header_and_filename(self, triage_setup):
        collection = ImageFileCollection(
            location=triage_setup.test_dir, keywords=["imagetyp"]
        )
        for header, fname in collection.headers(return_fname=True):
            assert fname in collection.summary["file"]
            assert isinstance(header, fits.Header)

    def test_dir_with_no_fits_files(self, tmpdir):
        empty_dir = tmpdir.mkdtemp()
        some_file = empty_dir.join("some_file.txt")
        some_file.dump("words")
        with pytest.warns(Warning) as w:
            collection = ImageFileCollection(
                location=empty_dir.strpath, keywords=["imagetyp"]
            )
        assert len(w) == 1
        assert str(w[0].message) == "no FITS files in the collection."
        assert collection.summary is None
        for _ in collection.headers():
            # This statement should not be reached if there are no FITS files
            assert 0

    def test_dir_with_no_keys(self, tmpdir):
        # This test should fail if the FITS files in the directory
        # are actually read.
        bad_dir = tmpdir.mkdtemp()
        not_really_fits = bad_dir.join("not_fits.fit")
        not_really_fits.dump("I am not really a FITS file")
        # Make sure an error will be generated if the FITS file is read
        with pytest.raises(IOError):
            fits.getheader(not_really_fits.strpath)

        log = tmpdir.join("tmp.log")
        self._setup_logger(log.strpath)

        _ = ImageFileCollection(location=bad_dir.strpath, keywords=[])

        with open(log.strpath) as f:
            warnings = f.read()

        # ImageFileCollection will suppress the IOError but log a warning
        # so check that the log has no warnings in it.
        assert len(warnings) == 0

    def test_fits_summary_when_keywords_are_not_subset(self, triage_setup):
        """
        Catch case when there is overlap between keyword list
        passed to the ImageFileCollection and to files_filtered
        but the latter is not a subset of the former.
        """
        ic = ImageFileCollection(
            triage_setup.test_dir, keywords=["imagetyp", "exposure"]
        )
        n_files = len(ic.files)
        files_missing_this_key = ic.files_filtered(imagetyp="*", monkeys=None)
        assert n_files > 0
        assert n_files == len(files_missing_this_key)

    def test_duplicate_keywords_in_setting(self, triage_setup):
        keywords_in = ["imagetyp", "a", "a"]
        ic = ImageFileCollection(triage_setup.test_dir, keywords=keywords_in)
        for key in set(keywords_in):
            assert key in ic.keywords
        # One keyword gets added: file
        assert len(ic.keywords) < len(keywords_in) + 1

    def test_keyword_includes_file(self, triage_setup):
        keywords_in = ["file", "imagetyp"]
        ic = ImageFileCollection(triage_setup.test_dir, keywords=keywords_in)
        assert "file" in ic.keywords
        file_keywords = [key for key in ic.keywords if key == "file"]
        assert len(file_keywords) == 1

    def test_setting_keywords_to_none(self, triage_setup):
        ic = ImageFileCollection(triage_setup.test_dir, keywords=["imagetyp"])
        ic.keywords = None
        assert ic.summary == []

    def test_getting_value_for_keyword(self, triage_setup):
        ic = ImageFileCollection(triage_setup.test_dir, keywords=["imagetyp"])
        # Does it fail if the keyword is not in the summary?
        with pytest.raises(ValueError):
            ic.values("filter")
        # If I ask for unique values do I get them?
        values = ic.values("imagetyp", unique=True)

        assert values == list(set(ic.summary["imagetyp"]))
        assert len(values) < len(ic.summary["imagetyp"])
        # Does the list of non-unique values match the raw column?
        values = ic.values("imagetyp", unique=False)
        assert values == list(ic.summary["imagetyp"])
        # Does unique actually default to false?
        values2 = ic.values("imagetyp")
        assert values == values2

    def test_collection_when_one_file_not_fits(self, triage_setup):
        not_fits = "foo.fit"
        path_bad = os.path.join(triage_setup.test_dir, not_fits)
        # Create an empty file...
        with open(path_bad, "w"):
            pass
        ic = ImageFileCollection(triage_setup.test_dir, keywords=["imagetyp"])
        assert not_fits not in ic.summary["file"]
        os.remove(path_bad)

    def test_data_type_mismatch_in_fits_keyword_values(self, triage_setup):
        # If one keyword has an unexpected type, do we notice?
        img = np.uint16(np.arange(100))
        bad_filter = fits.PrimaryHDU(img)
        bad_filter.header["imagetyp"] = "LIGHT"
        bad_filter.header["filter"] = 15.0
        path_bad = os.path.join(triage_setup.test_dir, "bad_filter.fit")
        bad_filter.writeto(path_bad)
        ic = ImageFileCollection(triage_setup.test_dir, keywords=["filter"])
        # dtype is object when there is a mix of types
        assert ic.summary["filter"].dtype == np.dtype("O")
        os.remove(path_bad)

    def test_filter_by_numerical_value(self, triage_setup):
        ic = ImageFileCollection(triage_setup.test_dir, keywords=["naxis"])
        should_be_zero = ic.files_filtered(naxis=2)
        assert len(should_be_zero) == 0
        should_not_be_zero = ic.files_filtered(naxis=1)
        assert len(should_not_be_zero) == triage_setup.n_test["files"]

    def test_files_filtered_with_full_path(self, triage_setup):
        ic = ImageFileCollection(triage_setup.test_dir, keywords=["naxis"])
        files = ic.files_filtered(naxis=1, include_path=True)

        for f in files:
            assert f.startswith(triage_setup.test_dir)

    def test_unknown_generator_type_raises_error(self, triage_setup):
        ic = ImageFileCollection(triage_setup.test_dir, keywords=["naxis"])
        with pytest.raises(ValueError):
            for _ in ic._generator("not a real generator"):
                pass

    def test_setting_write_location_to_bad_dest_raises_error(
        self, tmpdir, triage_setup
    ):
        new_tmp = tmpdir.mkdtemp()
        bad_directory = new_tmp.join("foo")

        ic = ImageFileCollection(triage_setup.test_dir, keywords=["naxis"])
        with pytest.raises(IOError):
            for _ in ic.headers(save_location=bad_directory.strpath):
                pass

    def test_initialization_with_no_keywords(self, triage_setup):
        # This test is primarily historical -- the old default for
        # keywords was an empty list (it is now the wildcard '*').
        ic = ImageFileCollection(location=triage_setup.test_dir, keywords=[])
        # Iteration below failed before bugfix...
        execs = 0
        for _ in ic.headers():
            execs += 1
        assert not execs

    def check_all_keywords_in_collection(self, image_collection):
        lower_case_columns = [c.lower() for c in image_collection.summary.colnames]
        for h in image_collection.headers():
            for k in h:
                assert k.lower() in lower_case_columns

    def test_tabulate_all_keywords(self, triage_setup):
        ic = ImageFileCollection(location=triage_setup.test_dir, keywords="*")
        self.check_all_keywords_in_collection(ic)

    def test_summary_table_is_always_masked(self, triage_setup):
        # First, try grabbing all of the keywords
        ic = ImageFileCollection(location=triage_setup.test_dir, keywords="*")
        assert ic.summary.masked
        # Now, try keywords that every file will have
        ic.keywords = ["bitpix"]
        assert ic.summary.masked
        # What about keywords that include some that will surely be missing?
        ic.keywords = ["bitpix", "dsafui"]
        assert ic.summary.masked

    def test_case_of_keywords_respected(self, triage_setup):
        keywords_in = ["BitPix", "instrume", "NAXIS"]
        ic = ImageFileCollection(location=triage_setup.test_dir, keywords=keywords_in)
        for key in keywords_in:
            assert key in ic.summary.colnames

    def test_grabbing_all_keywords_and_specific_keywords(self, triage_setup):
        keyword_not_in_headers = "OIdn89!@"
        ic = ImageFileCollection(
            triage_setup.test_dir, keywords=["*", keyword_not_in_headers]
        )
        assert keyword_not_in_headers in ic.summary.colnames
        self.check_all_keywords_in_collection(ic)

    def test_grabbing_all_keywords_excludes_empty_key(self, triage_setup):
        # This test needs a file with a blank keyword in it to ensure
        # that case is handled correctly.
        blank_keyword = fits.PrimaryHDU()
        blank_keyword.data = np.zeros((100, 100))
        blank_keyword.header[""] = "blank"

        blank_keyword.writeto(os.path.join(triage_setup.test_dir, "blank.fits"))

        ic = ImageFileCollection(triage_setup.test_dir, keywords="*")
        assert "col0" not in ic.summary.colnames

    @pytest.mark.skipif(
        "os.environ.get('APPVEYOR') or os.sys.platform == 'win32'",
        reason="fails on Windows because of file permissions.",
    )
    def test_header_with_long_history_roundtrips_to_disk(self, triage_setup):
        # I tried combing several history comments into one table entry with
        # '\n'.join(history), which resulted in a table that couldn't
        # round trip to disk because on read the newline character was
        # interpreted as...a new line! This test is a check against future
        # foolishness.
        from astropy.table import Table

        img = np.uint16(np.arange(100))
        long_history = fits.PrimaryHDU(img)
        long_history.header["imagetyp"] = "BIAS"
        long_history.header["history"] = "Something happened"
        long_history.header["history"] = "Then something else happened"
        long_history.header["history"] = "And then something odd happened"
        path_history = os.path.join(triage_setup.test_dir, "long_history.fit")
        long_history.writeto(path_history)
        ic = ImageFileCollection(triage_setup.test_dir, keywords="*")
        with NamedTemporaryFile() as test_table:
            ic.summary.write(test_table.name, format="ascii.csv", overwrite=True)
            table_disk = Table.read(test_table.name, format="ascii.csv")
        assert len(table_disk) == len(ic.summary)

    @pytest.mark.skipif(
        "os.environ.get('APPVEYOR') or os.sys.platform == 'win32'",
        reason="fails on Windows because file " "overwriting fails",
    )
    def test_refresh_method_sees_added_keywords(self, triage_setup):
        ic = ImageFileCollection(triage_setup.test_dir, keywords="*")
        # Add a keyword I know isn't already in the header to each file.
        not_in_header = "BARKARK"

        for h in ic.headers(overwrite=True):
            h[not_in_header] = True

        assert not_in_header not in ic.summary.colnames

        ic.refresh()
        # After refreshing the odd keyword should be present.
        assert not_in_header.lower() in ic.summary.colnames

    def test_refresh_method_sees_added_files(self, triage_setup):
        ic = ImageFileCollection(triage_setup.test_dir, keywords="*")
        # Compressed files don't get copied. Not sure why...
        original_len = len(ic.summary) - triage_setup.n_test["compressed"]
        # Generate additional files in this directory
        for _ in ic.headers(save_with_name="_foo"):
            pass
        ic.refresh()
        new_len = len(ic.summary) - triage_setup.n_test["compressed"]
        assert new_len == 2 * original_len

    def test_keyword_order_is_preserved(self, triage_setup):
        keywords = ["imagetyp", "exposure", "filter"]
        ic = ImageFileCollection(triage_setup.test_dir, keywords=keywords)
        assert ic.keywords == ["file"] + keywords

    def test_sorting(self, triage_setup):
        collection = ImageFileCollection(
            location=triage_setup.test_dir, keywords=["imagetyp", "filter", "object"]
        )

        all_elements = []
        for hdu, fname in collection.hdus(return_fname=True):
            all_elements.append((str(hdu.header), fname))
        # Now sort
        collection.sort(keys=["imagetyp", "object"])
        # and check it's all still right
        for hdu, fname in collection.hdus(return_fname=True):
            assert (str(hdu.header), fname) in all_elements
        for i in range(len(collection.summary)):
            assert collection.summary["file"][i] == collection.files[i]

    def test_sorting_without_key_fails(self, triage_setup):
        ic = ImageFileCollection(location=triage_setup.test_dir)
        with pytest.raises(ValueError):
            ic.sort(keys=None)

    def test_duplicate_keywords(self, triage_setup):
        # Make sure duplicated keywords don't make the imagefilecollection
        # fail.
        hdu = fits.PrimaryHDU()
        hdu.data = np.zeros((5, 5))
        hdu.header["stupid"] = "fun"
        hdu.header.append(("stupid", "nofun"))

        hdu.writeto(os.path.join(triage_setup.test_dir, "duplicated.fits"))

        with pytest.warns(UserWarning) as w:
            ic = ImageFileCollection(triage_setup.test_dir, keywords="*")
        assert len(w) == 1
        assert "stupid" in str(w[0].message)

        assert "stupid" in ic.summary.colnames
        assert "fun" in ic.summary["stupid"]
        assert "nofun" not in ic.summary["stupid"]

    def test_ccds_generator_in_different_directory(self, triage_setup, tmpdir):
        """
        Regression test for https://github.com/astropy/ccdproc/issues/421 in
        which the ccds generator fails if the current working directory is
        not the location of the ImageFileCollection.
        """

        coll = ImageFileCollection(triage_setup.test_dir)

        # The temporary directory below should be different that the collection
        # location.
        os.chdir(tmpdir.strpath)

        # Let's make sure it is.
        assert not os.path.samefile(os.getcwd(), coll.location)

        # This generated an IOError before the issue was fixed.
        for _ in coll.ccds(ccd_kwargs={"unit": "adu"}):
            pass

    def test_ccds_generator_does_not_support_overwrite(self, triage_setup):
        """
        CCDData objects have several attributes that make it hard to
        reliably support overwriting. For example in what extension should
        mask, uncertainty be written?
        Also CCDData doesn't explicitly support in-place operations so it's to
        easy to create a new CCDData object inadvertantly and all modifications
        might be lost.
        """
        ic = ImageFileCollection(triage_setup.test_dir)
        with pytest.raises(NotImplementedError):
            ic.ccds(overwrite=True)
        with pytest.deprecated_call(), pytest.raises(NotImplementedError):
            ic.ccds(clobber=True)

    def test_glob_matching(self, triage_setup):
        # We'll create two files with strange names to test glob
        # includes / excludes
        one = fits.PrimaryHDU()
        one.data = np.zeros((5, 5))
        one.header[""] = "whatever"

        one.writeto(os.path.join(triage_setup.test_dir, "SPAM_stuff.fits"))
        one.writeto(os.path.join(triage_setup.test_dir, "SPAM_other_stuff.fits"))

        coll = ImageFileCollection(triage_setup.test_dir, glob_include="SPAM*")
        assert len(coll.files) == 2

        coll = ImageFileCollection(
            triage_setup.test_dir, glob_include="SPAM*", glob_exclude="*other*"
        )
        assert len(coll.files) == 1

        # The glob attributes are readonly, so setting them raises an Exception.
        with pytest.raises(AttributeError):
            coll.glob_exclude = "*stuff*"
        with pytest.raises(AttributeError):
            coll.glob_include = "*stuff*"

    def test_that_test_files_have_expected_properties(self, triage_setup):
        expected_name = get_pkg_data_filename("data/expected_ifc_file_properties.csv")
        expected = Table.read(expected_name)

        # Make the comparison more reliable by sorting
        expected.sort("file")
        ic = ImageFileCollection(triage_setup.test_dir)
        actual = ic.summary
        # Write the actual IFC summary out to disk to turn bool into strings of
        # "True" and "False", and any other non-essential differences between
        # the tables.
        tmp_file = "actual.csv"
        actual.write(tmp_file)
        actual = Table.read(tmp_file)

        # Make the comparison more reliable by sorting
        actual.sort("file")
        assert len(ic.summary) == len(expected)

        for column in expected.colnames:
            assert np.all(actual[column] == expected[column])

    def test_image_collection_with_no_location(self, triage_setup):
        # Test for a feature requested in
        #
        #     https://github.com/astropy/ccdproc/issues/374
        #
        # and fix the bug reported in
        #
        #     https://github.com/astropy/ccdproc/issues/662
        #
        # Create a collection from a list of file names (which can include
        # path as needed)

        source_path = Path(triage_setup.test_dir)

        # Put the first three files in source_path into temp_path below
        # then create the image collection out of the three in temp_path and
        # the rest in source_path.
        source_files = [p for p in source_path.iterdir()]

        move_to_temp = source_files[:3]
        keep_in_source = source_files[3:]

        with TemporaryDirectory() as td:
            temp_dir = Path(td)
            file_paths = []
            for source in move_to_temp:
                temp_path = temp_dir / source.name
                temp_path.write_bytes(source.read_bytes())
                file_paths.append(str(temp_path))
            file_paths.extend(str(p) for p in keep_in_source)

            # Move up a level to make sure we are not accidentally
            # pulling in files from the current working directory,
            # which includes everythin in source.
            os.chdir("..")

            ic = ImageFileCollection(filenames=file_paths)
            assert len(ic.summary) == len(file_paths)

            expected_name = get_pkg_data_filename(
                "data/expected_ifc_file_properties.csv"
            )
            expected = Table.read(expected_name)

            # Make the comparison more reliable by sorting
            expected.sort("file")

            actual = ic.summary

            # Write the actual IFC summary out to disk to turn bool into strings
            # of"True" and "False", and any other non-essential differences
            # between the tables.
            tmp_file = "actual.csv"
            actual.write(tmp_file)
            actual = Table.read(tmp_file)
            # Make the comparison more reliable by sorting...but the actual
            # in this case includes paths, so we really want to sort by the
            # base name of the file.
            bases = np.array([Path(f).name for f in actual["file"]])
            sort_order = np.argsort(bases)
            actual = actual[sort_order]
            bases = bases[sort_order]

            assert all(Path(f).exists() for f in actual["file"])

            for column in expected.colnames:
                if column == "file":
                    assert np.all(bases == expected[column])
                else:
                    assert np.all(actual[column] == expected[column])

            # Set comparisons don't care about order :)
            # Check several of the ways we can get file names from the
            # collection and ensure all of them include the path.
            assert set(file_paths) == set(ic.summary["file"])
            assert set(file_paths) == set(ic.files)
            assert set(file_paths) == set(ic.files_filtered(include_path=True))

            # Spot check a couple of dtypes as a test for
            #   https://github.com/astropy/ccdproc/issues/662
            assert ic.summary["extend"].dtype == "bool"

            # Of course, default dtypes on Windows are different. So instead
            # of comparing to something sensible like int64, compare to the
            # default int dtype.
            assert ic.summary["naxis1"].dtype == np.array([5]).dtype

            # and the default float dtype
            assert ic.summary["exposure"].dtype == np.array([5.0]).dtype

            expected_heads = (actual["imagetyp"] == "LIGHT").sum()

            n_heads = 0
            # Try one of the generators
            for h in ic.headers(imagetyp="light"):
                assert h["imagetyp"].lower() == "light"
                n_heads += 1

            assert n_heads == expected_heads

    def test_force_detect_fits_files_finds_fits_files(self, triage_setup):
        # Tests for new feature
        #
        #     https://github.com/astropy/ccdproc/issues/620
        #
        # which supports adding all of the FITS files in a location based on
        # their *contents* instead of their *extension*.

        # Grab files from the default collection and make a copy with a new name
        # (and with no fits-like extension)
        #
        # Making a copy of *every* file means we can just double the expected
        # number of files as part of the tests.
        path = Path(triage_setup.test_dir)

        for idx, p in enumerate(path.iterdir()):
            new_name = f"no_extension{idx}"
            new_path = path / new_name
            new_path.write_bytes(p.read_bytes())

        ic = ImageFileCollection(location=str(path), find_fits_by_reading=True)

        # Compressed files won't be automagically detected by reading the
        # first few bytes.
        expected_number = (
            2 * triage_setup.n_test["files"] - triage_setup.n_test["compressed"]
        )

        assert len(ic.summary) == expected_number

        n_bias = (ic.summary["imagetyp"] == "BIAS").sum()
        assert n_bias == 2 * triage_setup.n_test["bias"]

        # Only one file in the original set of test files has exposure time
        # 15, so there should be two now.
        assert len(ic.files_filtered(exposure=15.0)) == 2

        # Try one of the generators
        expected_heads = (
            2 * triage_setup.n_test["light"] - triage_setup.n_test["compressed"]
        )

        n_heads = 0

        for h in ic.headers(imagetyp="light"):
            assert h["imagetyp"].lower() == "light"
            n_heads += 1

        assert n_heads == expected_heads

    @pytest.mark.filterwarnings(
        "ignore:The following header keyword is invalid:UserWarning"
    )
    def test_less_strict_verify_option(self, triage_setup):
        # Tests for feature request
        #
        #    https://github.com/astropy/ccdproc/issues/607
        #
        # which would allow reading of otherwise invalid FITS headers.

        bad_header = """
            NAXIS1  =                   10 / length of data axis 1
            NAXIS2  =                   10 / length of data axis 2
            TESTVERI= '2017/02/13-16:51:38 / Test VerifyWarning
        """

        with pytest.warns(VerifyWarning):
            testh = fits.Header.fromstring(bad_header)

            testfits = fits.PrimaryHDU(data=np.ones((10, 10)), header=testh)

            path = Path(triage_setup.test_dir)
            bad_fits_name = "test_warnA.fits"
            testfits.writeto(path / bad_fits_name, output_verify="warn", overwrite=True)

            ic = ImageFileCollection(location=str(path))

        assert bad_fits_name in ic.files

        # Turns out this sample header is so messed up that TESTVERI does not
        # end up as a keyword.
        assert "TESTVERI" not in ic.summary.colnames

        # This does end up as a key some how *shrug*.
        assert "17/02/13" in ic.summary.colnames

        # Try making the summary as in the original bug report
        with pytest.warns(
            AstropyUserWarning, match="The following header keyword is invalid"
        ):
            ic = ImageFileCollection(location=str(path), glob_include="*warnA*")

    def test_type_of_empty_collection(self, triage_setup):
        # Test for implementation of the suggestion in
        #
        #    https://github.com/astropy/ccdproc/issues/601
        #
        # in which an empty collection with no keys has but with files
        # returns a summary table with one column, but an empty collection
        # with no keys and no files returns None.

        # Make a dummy keyword that we then delete.
        ic = ImageFileCollection(triage_setup.test_dir, keywords=["fafa"])
        ic.keywords = []
        assert set(ic.summary.colnames) == set(["file"])

        # Remove all of the fits files
        path = Path(triage_setup.test_dir)
        for p in path.iterdir():
            p.unlink()

        # Now the summary should be none
        with pytest.warns(AstropyUserWarning, match="no FITS files in the collection"):
            ic = ImageFileCollection(triage_setup.test_dir)
        assert ic.summary is None
        assert ic.keywords == []
        assert ic.files_filtered() == []

    def test_regex_match_for_search(self, triage_setup):
        # Test regex matching in searches

        ic = ImageFileCollection(triage_setup.test_dir)

        files = ic.files_filtered(regex_match=True, imagetyp="b.*s")
        assert len(files) == triage_setup.n_test["bias"]

        # This should return all of the files in the test set
        all_files = ic.files_filtered(regex_match=True, imagetyp="bias|light")
        assert len(all_files) == triage_setup.n_test["files"]

        # Add a column with more interesting content and see whether we
        # match that.
        ic.summary["match_me"] = [
            "hello",
            "goodbye",
            "bye",
            "byte",
            "good bye hello",
            "dog",
        ]

        hello_anywhere = ic.files_filtered(regex_match=True, match_me="hello")
        assert len(hello_anywhere) == 2

        hello_start = ic.files_filtered(regex_match=True, match_me="^hello")
        assert len(hello_start) == 1

        # Is it really a case-insensitive match?
        hello_start = ic.files_filtered(regex_match=True, match_me="^HeLlo")
        assert len(hello_start) == 1

        any_bye = ic.files_filtered(regex_match=True, match_me="by.*e")
        assert len(any_bye) == 4

    def test_generator_with_regex(self, triage_setup):
        ic = ImageFileCollection(triage_setup.test_dir)

        n_light = 0

        for h in ic.headers(regex_match=True, imagetyp="li.*t"):
            assert h["imagetyp"].lower() == "light"
            n_light += 1

        assert n_light == triage_setup.n_test["light"]

    def test_make_collection_by_filtering(self, triage_setup):
        # Test for implementation of feature at
        #
        #    https://github.com/astropy/ccdproc/issues/596
        #
        # which adds the ability to create a new collection by filtering
        # an existing ImageFileCollection.

        ic = ImageFileCollection(location=triage_setup.test_dir)
        new_ic = ic.filter(imagetyp="light")
        assert len(new_ic.summary) == triage_setup.n_test["light"]
        for header in new_ic.headers():
            assert header["imagetyp"].lower() == "light"

    def test_filtered_collection_with_no_files(self, triage_setup):
        ifc = ImageFileCollection(triage_setup.test_dir)

        with pytest.warns(AstropyUserWarning, match="no FITS files"):
            _ = ifc.filter(object="really fake object")

    def test_filter_on_regex_escape_characters(self, triage_setup):
        # Test for implementation of bugfix at
        #
        #    https://github.com/astropy/ccdproc/issues/770
        #
        # which escapes regex special characters in keyword values used for
        # filtering a collection, when option `regex_match=False`.

        # For a few different special characters, make test files with FILTER
        # keyword containing these

        special_kwds = ["CO+", "GG420 (1)", "V|R|I", "O[III]", "NaD^2"]
        for i, kw in enumerate(special_kwds, 1):
            hdu = fits.PrimaryHDU()
            hdu.data = np.zeros((5, 5))
            hdu.header["REGEX_FL"] = kw
            hdu.writeto(
                os.path.join(triage_setup.test_dir, f"regex_special_{i:d}.fits")
            )

        ic = ImageFileCollection(triage_setup.test_dir)
        for kw in special_kwds:
            new_ic = ic.filter(regex_fl=kw)
            assert len(new_ic.files) == 1
            assert kw in new_ic.summary["regex_fl"]



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np
import pytest

from ..slices import slice_from_string


# none of these are properly enclosed in brackets; is an error raised?
@pytest.mark.parametrize("arg", ["1:2", "[1:2", "1:2]"])
def test_slice_from_string_needs_enclosing_brackets(arg):
    with pytest.raises(ValueError):
        slice_from_string(arg)


@pytest.mark.parametrize(
    "start,stop,step",
    [
        (None, None, -1),
        (5, 10, None),
        (None, 25, None),
        (2, 30, 3),
        (30, None, -2),
        (None, None, None),
    ],
)
def test_slice_from_string_1d(start, stop, step):
    an_array = np.zeros([100])

    def stringify(n):
        return str(n) if n else ""

    start_str = stringify(start)
    stop_str = stringify(stop)
    step_str = stringify(step)

    if step_str:
        slice_str = ":".join([start_str, stop_str, step_str])
    else:
        slice_str = ":".join([start_str, stop_str])
    sli = slice_from_string("[" + slice_str + "]")
    expected = an_array[slice(start, stop, step)]
    np.testing.assert_allclose(expected, an_array[sli])


@pytest.mark.parametrize("arg", ["  [ 1:  45]", "[ 1  :4 5]", "  [1:45] "])
def test_slice_from_string_spaces(arg):
    an_array = np.zeros([100])
    np.testing.assert_allclose(an_array[1:45], an_array[slice_from_string(arg)])


def test_slice_from_string_2d():
    an_array = np.zeros([100, 200])

    # manually writing a few cases here rather than parametrizing because the
    # latter seems not worth the trouble.
    sli = slice_from_string("[:-1:2, :]")
    np.testing.assert_allclose(an_array[:-1:2, :], an_array[sli])

    sli = slice_from_string("[:, 15:90]")
    np.testing.assert_allclose(an_array[:, 15:90], an_array[sli])

    sli = slice_from_string("[10:80:5, 15:90:-1]")
    np.testing.assert_allclose(an_array[10:80:5, 15:90:-1], an_array[sli])


def test_slice_from_string_fits_style():
    sli = slice_from_string("[1:5, :]", fits_convention=True)
    # order is reversed, so is the *first* slice one that includes everything?
    assert sli[0].start is None and sli[0].stop is None and sli[0].step is None
    # In the second slice, has the first index been reduced by 1 and the
    # second index left unchanged?
    assert sli[1].start == 0 and sli[1].stop == 5
    sli = slice_from_string("[1:10:2, 4:5:2]", fits_convention=True)
    assert sli[0] == slice(3, 5, 2)
    assert sli[1] == slice(0, 10, 2)


def test_slice_from_string_fits_inverted():
    sli = slice_from_string("[20:10:2, 10:5, 5:4]", fits_convention=True)
    assert sli[0] == slice(4, 2, -1)
    assert sli[1] == slice(9, 3, -1)
    assert sli[2] == slice(19, 8, -2)
    # Handle a bunch of special cases for inverted slices, when the
    # stop index is 1 or 2
    sli = slice_from_string("[20:1:4, 21:1:4, 22:2:4, 2:1]", fits_convention=True)
    assert sli[0] == slice(1, None, -1)
    assert sli[1] == slice(21, 0, -4)
    assert sli[2] == slice(20, None, -4)
    assert sli[3] == slice(19, None, -4)


def test_slice_from_string_empty():
    assert len(slice_from_string("")) == 0


def test_slice_from_string_bad_fits_slice():
    with pytest.raises(ValueError):
        # Do I error because 0 is an illegal lower bound?
        slice_from_string("[0:10, 1:5]", fits_convention=True)
    with pytest.raises(ValueError):
        # Same as above, but switched order
        slice_from_string("[1:5, 0:10]", fits_convention=True)
    with pytest.raises(ValueError):
        # Do I error if an ending index is negative?
        slice_from_string("[1:10, 10:-1]", fits_convention=True)


def test_slice_from_string_fits_wildcard():
    sli = slice_from_string("[*,-*]", fits_convention=True)
    assert sli[0] == slice(None, None, -1)
    assert sli[1] == slice(None, None, None)
    sli = slice_from_string("[*:2,-*:2]", fits_convention=True)
    assert sli[0] == slice(None, None, -2)
    assert sli[1] == slice(None, None, 2)



import os
import subprocess
import sys
from pathlib import Path

import pytest

run_dir = Path(__file__).parent

# Why? So that we get up to the file above ccdproc, so that in the
# subprocess we can add that direction to sys.path.
subprocess_dir = run_dir.parent.parent

OVERHEAD = "4"
NUM_FILE_LIMIT = "20"
common_args = [
    sys.executable,
    str(run_dir / "run_with_file_number_limit.py"),
    "--kind",
    "fits",
    "--overhead",
    OVERHEAD,
]


# Regression test for #629
@pytest.mark.skipif(
    os.environ.get("APPVEYOR") or os.sys.platform == "win32",
    reason="Test relies on linux/osx features of psutil",
)
def test_open_files_combine_no_chunks():
    """
    Test that we are not opening (much) more than the number of files
    we are processing.
    """
    # Make a copy
    args = list(common_args)
    args.extend(["--open-by", "combine-nochunk", NUM_FILE_LIMIT])
    p = subprocess.run(args=args, cwd=str(subprocess_dir))
    # If we have succeeded the test passes. We are only checking that
    # we don't have too many files open.
    assert p.returncode == 0


# Regression test for #629
@pytest.mark.skipif(
    os.environ.get("APPVEYOR") or os.sys.platform == "win32",
    reason="Test relies on linux/osx features of psutil",
)
def test_open_files_combine_chunks():
    """
    Test that we are not opening (much) more than the number of files
    we are processing when combination is broken into chunks.
    """
    # Make a copy
    args = list(common_args)
    args.extend(["--open-by", "combine-chunk", NUM_FILE_LIMIT])
    p = subprocess.run(args=args, cwd=str(subprocess_dir))
    # If we have succeeded the test passes. We are only checking that
    # we don't have too many files open.
    assert p.returncode == 0
